"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

import pathlib
from datetime import datetime, timezone
from itertools import product
from typing import Dict, List

import ee
import numpy as np
import pytest
import rasterio as rio
from rasterio import Affine, features, warp, windows

from geedim import utils
from geedim.download import _nodata_vals, BaseImage
from geedim.enums import ExportType, ResamplingMethod
from tests.conftest import region_25ha


class BaseImageLike:
    """Mock ``BaseImage`` for ``_get_tile_shape()`` and ``tiles()``."""

    def __init__(
        self, shape: tuple[int, int], count: int = 10, dtype: str = 'uint16', transform: Affine = Affine.identity()
    ):
        self.shape = shape
        self.count = count
        self.dtype = dtype
        self.transform = transform
        dtype_size = np.dtype(dtype).itemsize
        self.size = shape[0] * shape[1] * count * dtype_size

    _tiles = BaseImage._tiles
    _get_tile_shape = BaseImage._get_tile_shape


def _bounds(geom: dict, dst_crs: str | rio.CRS = 'EPSG:4326') -> tuple[float, ...]:
    """Return the bounds of GeoJSON polygon ``geom`` in the ``dst_crs`` CRS."""
    # transform geom, then find bounds (if geom is an ee footprint, finding its bounds then transforming expands
    # bounds beyond their true location)
    src_crs = utils.rio_crs(geom['crs']['properties']['name']) if 'crs' in geom else 'EPSG:4326'
    dst_crs = utils.rio_crs(dst_crs) or 'EPSG:4326'
    geom = geom if dst_crs == src_crs else warp.transform_geom(src_crs, dst_crs, geom)
    return features.bounds(geom)


def _intersect_bounds(bounds1: tuple[float, ...], bounds2: tuple[float, ...]) -> tuple[float, ...] | None:
    """Return the intersection of ``bounds1`` and ``bounds2``, or ``None`` when there is no intersection."""
    bounds = np.array([*np.fmax(bounds1[:2], bounds2[:2]), *np.fmin(bounds1[2:], bounds2[2:])])
    return tuple(bounds.tolist()) if np.all((bounds[2:] - bounds[:2]) > 0) else None


@pytest.fixture(scope='session')
def user_base_image() -> BaseImage:
    """A BaseImage instance where the encapsulated image has no fixed projection or ID."""
    return BaseImage(ee.Image([1, 2, 3]))


@pytest.fixture(scope='session')
def user_fix_base_image() -> BaseImage:
    """A BaseImage instance where the encapsulated image has a fixed projection (EPSG:4326), a scale in degrees,
    is unbounded and has no ID.
    """
    return BaseImage(ee.Image([1, 2, 3]).setDefaultProjection(crs='EPSG:4326', scale=30))


@pytest.fixture(scope='session')
def user_fix_bnd_base_image(region_10000ha) -> BaseImage:
    """A BaseImage instance where the encapsulated image has a fixed projection (EPSG:4326), a scale in degrees,
    is bounded, and has no ID.
    """
    return BaseImage(
        ee.Image([1, 2, 3]).setDefaultProjection(crs='EPSG:4326', scale=30).clipToBoundsAndScale(region_10000ha)
    )


@pytest.fixture(scope='session')
def s2_sr_hm_base_image(s2_sr_hm_image_id: str) -> BaseImage:
    """A BaseImage instance encapsulating a Sentinel-2 image.  Covers `region_*ha`."""
    return BaseImage.from_id(s2_sr_hm_image_id)


@pytest.fixture(scope='session')
def l9_base_image(l9_image_id: str) -> BaseImage:
    """A BaseImage instance encapsulating a Landsat-9 image.  Covers `region_*ha`."""
    return BaseImage.from_id(l9_image_id)


@pytest.fixture(scope='session')
def l8_base_image(l8_image_id: str) -> BaseImage:
    """A BaseImage instance encapsulating a Landsat-8 image.  Covers `region_*ha`."""
    return BaseImage.from_id(l8_image_id)


@pytest.fixture(scope='session')
def landsat_ndvi_base_image(landsat_ndvi_image_id: str) -> BaseImage:
    """A BaseImage instance encapsulating a Landsat NDVI composite image.  Covers `region_*ha`."""
    return BaseImage.from_id(landsat_ndvi_image_id)


@pytest.fixture(scope='session')
def modis_nbar_base_image_unbnd(modis_nbar_image_id: str) -> BaseImage:
    """A BaseImage instance encapsulating an unbounded MODIS NBAR image in its native CRS.  Covers `region_*ha`."""
    return BaseImage(ee.Image(modis_nbar_image_id))


@pytest.fixture(scope='session')
def modis_nbar_base_image(modis_nbar_image_id: str, region_100ha: Dict) -> BaseImage:
    """A BaseImage instance encapsulating a MODIS NBAR image.  Covers `region_*ha`."""
    return BaseImage(ee.Image(modis_nbar_image_id).clipToBoundsAndScale(region_100ha))


@pytest.fixture(scope='session')
def google_dyn_world_base_image(google_dyn_world_image_id: str) -> BaseImage:
    """A BaseImage instance encapsulating a Google Dynamic World image with positive y-axis transform.  Covers
    `region_*ha`.
    """
    return BaseImage.from_id(google_dyn_world_image_id)


def bounds_polygon(left: float, bottom: float, right: float, top: float, crs: str = None):
    """Return a geojson polygon of the given bounds."""
    coordinates = [
        [
            [left, bottom],
            [left, top],
            [right, top],
            [right, bottom],
            [left, bottom],
        ]
    ]
    poly = dict(type='Polygon', coordinates=coordinates)
    if crs and crs != 'EPSG:4326':
        poly.update(crs=dict(type='name', properties=dict(name=crs)))
    return poly


def _test_export_image(exp_image: BaseImage, ref_image: BaseImage, **exp_kwargs):
    """Test ``exp_image`` against ``ref_image``, and optional crs, region & scale ``exp_kwargs`` arguments to
    ``BaseImage._prepare_for_export``.
    """
    # remove None value items from exp_kwargs
    exp_kwargs = {k: v for k, v in exp_kwargs.items() if v is not None}

    assert exp_image.dtype == exp_kwargs.get('dtype', ref_image.dtype)
    assert exp_image.band_properties == ref_image.band_properties

    # test crs and scale
    crs = exp_kwargs.get('crs', ref_image.crs)
    assert exp_image.crs == crs

    if 'shape' in exp_kwargs:
        assert exp_image.shape == exp_kwargs['shape']
    else:
        ref_scale = exp_kwargs.get('scale', ref_image.scale)
        assert exp_image.scale == ref_scale

    # test export bounds contain reference bounds
    region = exp_kwargs.get('region', ref_image.footprint)
    ref_bounds = _bounds(region, exp_image.crs)
    exp_bounds = _bounds(exp_image.footprint, exp_image.crs)
    tol = 1e-9 if exp_image.crs == 'EPSG:4326' else 1e-6
    assert _intersect_bounds(exp_bounds, ref_bounds) == pytest.approx(ref_bounds, abs=tol)

    # test export transform is on the reference grid
    if {'crs', 'scale', 'shape'}.isdisjoint(exp_kwargs.keys()):
        ji = ~exp_image.transform * (ref_image.transform[2], ref_image.transform[5])
        assert ji == pytest.approx(np.round(ji), abs=1e-6)


def test_id_name(user_base_image: BaseImage, s2_sr_hm_base_image: BaseImage):
    """Test ``id`` and ``name`` properties for different scenarios."""
    assert user_base_image.id is None
    assert user_base_image.name is None
    # check that BaseImage.from_id() sets id without a getInfo
    assert s2_sr_hm_base_image._id is not None
    assert s2_sr_hm_base_image.name == s2_sr_hm_base_image.id.replace('/', '-')


def test_user_props(user_base_image: BaseImage):
    """Test non fixed projection image properties (other than ``id`` and ``has_fixed_projection``)."""
    assert user_base_image.crs is None
    assert user_base_image.scale is None
    assert user_base_image.transform is None
    assert user_base_image.shape is None
    assert user_base_image.date is None
    assert user_base_image.size is None
    assert user_base_image.footprint is None
    assert user_base_image.dtype == 'uint8'
    assert user_base_image.count == 3


def test_fix_user_props(user_fix_base_image: BaseImage):
    """Test fixed projection image properties (other than ``id`` and ``has_fixed_projection``)."""
    assert user_fix_base_image.crs == 'EPSG:4326'
    assert user_fix_base_image.scale == pytest.approx(30, abs=1e-6)
    assert user_fix_base_image.transform is not None
    assert user_fix_base_image.shape is None
    assert user_fix_base_image.date is None
    assert user_fix_base_image.size is None
    assert user_fix_base_image.footprint is None
    assert user_fix_base_image.dtype == 'uint8'
    assert user_fix_base_image.count == 3


def test_s2_props(s2_sr_hm_base_image: BaseImage):
    """Test fixed projection S2 image properties (other than ``id`` and ``has_fixed_projection``)."""
    min_band_info = s2_sr_hm_base_image._ee_info['bands'][1]
    assert s2_sr_hm_base_image.crs == min_band_info['crs']
    assert s2_sr_hm_base_image.scale == min_band_info['crs_transform'][0]
    assert s2_sr_hm_base_image.transform == Affine(*min_band_info['crs_transform'])
    assert s2_sr_hm_base_image.shape == tuple(min_band_info['dimensions'][::-1])
    assert s2_sr_hm_base_image.date == datetime.fromtimestamp(
        s2_sr_hm_base_image.properties['system:time_start'] / 1000, timezone.utc
    )
    assert s2_sr_hm_base_image.size is not None
    assert s2_sr_hm_base_image.footprint is not None
    assert s2_sr_hm_base_image.footprint['type'] == 'Polygon'
    assert s2_sr_hm_base_image.dtype == 'uint32'
    assert s2_sr_hm_base_image.count == len(s2_sr_hm_base_image._ee_info['bands'])


@pytest.mark.parametrize(
    'base_image', ['landsat_ndvi_base_image', 's2_sr_hm_base_image', 'l9_base_image', 'modis_nbar_base_image']
)
def test_band_props(base_image: str, request: pytest.FixtureRequest):
    """Test ``band_properties`` completeness for generic / user / reflectance images."""
    base_image: BaseImage = request.getfixturevalue(base_image)
    assert base_image.band_properties is not None
    assert [bd['name'] for bd in base_image.band_properties] == [bd['id'] for bd in base_image._ee_info['bands']]
    for key in ['gsd', 'description']:
        has_key = [key in bd for bd in base_image.band_properties]
        assert all(has_key)


def test_has_fixed_projection(user_base_image: BaseImage, user_fix_base_image: BaseImage, s2_sr_hm_base_image):
    """Test the ``has_fixed_projection`` property."""
    assert not user_base_image.has_fixed_projection
    assert user_fix_base_image.has_fixed_projection
    assert s2_sr_hm_base_image.has_fixed_projection


@pytest.mark.parametrize(
    'ee_data_type_list, exp_dtype',
    [
        ([{'precision': 'int', 'min': 10, 'max': 11}, {'precision': 'int', 'min': 100, 'max': 101}], 'uint8'),
        ([{'precision': 'int', 'min': -128, 'max': -100}, {'precision': 'int', 'min': 0, 'max': 127}], 'int8'),
        ([{'precision': 'int', 'min': 256, 'max': 257}], 'uint16'),
        ([{'precision': 'int', 'min': -32768, 'max': 32767}], 'int16'),
        ([{'precision': 'int', 'min': 2**15, 'max': 2**32 - 1}], 'uint32'),
        ([{'precision': 'int', 'min': -(2**31), 'max': 2**31 - 1}], 'int32'),
        ([{'precision': 'float', 'min': 0.0, 'max': 1.0e9}, {'precision': 'float', 'min': 0.0, 'max': 1.0}], 'float32'),
        (
            [{'precision': 'int', 'min': 0.0, 'max': 2**31 - 1}, {'precision': 'float', 'min': 0.0, 'max': 1.0}],
            'float64',
        ),
        ([{'precision': 'int', 'min': 0, 'max': 255}, {'precision': 'double', 'min': -1e100, 'max': 1e100}], 'float64'),
    ],
)
def test_min_dtype(ee_data_type_list: List, exp_dtype: str):
    """Test ``BasicImage.__get_min_dtype()`` with emulated EE info dicts."""
    ee_info = dict(bands=[])
    for ee_data_type in ee_data_type_list:
        ee_info['bands'].append(dict(data_type=ee_data_type))
    assert BaseImage._get_min_dtype(ee_info) == exp_dtype


def test_convert_dtype_error():
    """Test ``BaseImage.test_convert_dtype()`` raises an error with incorrect dtype."""
    with pytest.raises(TypeError):
        BaseImage._convert_dtype(ee.Image(1), dtype='unknown')


@pytest.mark.parametrize('size, exp_str', [(1024, '1.02 KB'), (234.56e6, '234.56 MB'), (1e9, '1.00 GB')])
def test_str_format_size(size: int, exp_str: str):
    """Test formatting of byte sizes as human readable strings."""
    assert BaseImage._str_format_size(size) == exp_str


@pytest.mark.parametrize(
    'params',
    [
        dict(),
        dict(crs='EPSG:4326'),
        dict(crs='EPSG:4326', region='region_25ha'),
        dict(crs='EPSG:4326', scale=100),
        dict(crs='EPSG:4326', crs_transform=Affine.identity()),
        dict(crs='EPSG:4326', shape=(100, 100)),
        dict(region='region_25ha', scale=100),
        dict(crs_transform=Affine.identity(), shape=(100, 100)),
    ],
)
def test_prepare_no_fixed_projection(user_base_image: BaseImage, params: Dict, request: pytest.FixtureRequest):
    """Test ``BaseImage._prepare_for_export()`` raises an exception when the image has no fixed projection,
    and insufficient CRS & region defining arguments.
    """
    if 'region' in params:
        params['region'] = request.getfixturevalue(params['region'])
    with pytest.raises(ValueError) as ex:
        user_base_image._prepare_for_export(**params)
    assert 'does not have a fixed projection' in str(ex)


@pytest.mark.parametrize(
    'base_image, params',
    [
        ('user_fix_base_image', dict()),
        ('modis_nbar_base_image_unbnd', dict(crs='EPSG:4326')),
        ('modis_nbar_base_image_unbnd', dict(crs='EPSG:4326', scale=100)),
        ('modis_nbar_base_image_unbnd', dict(crs='EPSG:4326', crs_transform=Affine.identity())),
        ('modis_nbar_base_image_unbnd', dict(crs='EPSG:4326', shape=(100, 100))),
    ],
)
def test_prepare_unbounded(base_image: BaseImage, params: Dict, request: pytest.FixtureRequest):
    """Test ``BaseImage._prepare_for_export()`` raises an exception when the image is unbounded, and insufficient
    bounds defining parameters are specified.
    """
    base_image: BaseImage = request.getfixturevalue(base_image)
    with pytest.raises(ValueError) as ex:
        base_image._prepare_for_export(**params)
    assert 'This image is unbounded' in str(ex)


def test_prepare_exceptions(user_base_image: BaseImage, user_fix_base_image: BaseImage, region_25ha: Dict):
    """Test remaining ``BaseImage._prepare_for_export()`` error cases."""
    with pytest.raises(ValueError):
        # no fixed projection and resample
        user_base_image._prepare_for_export(
            region=region_25ha, crs='EPSG:3857', scale=30, resampling=ResamplingMethod.bilinear
        )


@pytest.mark.parametrize(
    'base_image',
    [
        'user_fix_bnd_base_image',
        's2_sr_hm_base_image',
        'l9_base_image',
        'modis_nbar_base_image',
        'google_dyn_world_base_image',
    ],
)
def test_prepare_defaults(base_image: str, request: pytest.FixtureRequest):
    """Test ``BaseImage._prepare_for_export()`` with no (i.e. default) arguments."""
    base_image: BaseImage = request.getfixturevalue(base_image)
    exp_image = base_image._prepare_for_export()

    _test_export_image(exp_image, base_image)


@pytest.mark.parametrize(
    'base_image, crs, crs_transform, shape',
    [
        ('user_base_image', 'EPSG:4326', (-1e-3, 0, -180.1, 0, 2e-3, 0.1), (300, 400)),
        ('user_fix_base_image', 'EPSG:32734', (10, 0, 100, 0, 10, 200), (300, 400)),
        ('s2_sr_hm_base_image', 'EPSG:3857', (10, 0, 100, 0, -10, 200), (300, 400)),
    ],
)
def test_prepare_transform_shape(
    base_image: str, crs: str, crs_transform: tuple[float, ...], shape: tuple[int, int], request: pytest.FixtureRequest
):
    """Test ``BaseImage._prepare_for_export()`` with ``crs_transform`` and ``shape`` parameters."""
    base_image: BaseImage = request.getfixturevalue(base_image)
    exp_image = base_image._prepare_for_export(crs=crs, crs_transform=crs_transform, shape=shape)

    assert exp_image.dtype == base_image.dtype
    assert exp_image.band_properties == base_image.band_properties

    assert exp_image.crs == crs or base_image.crs
    # assert exp_image.scale == np.abs((crs_transform[0], crs_transform[4])).mean()
    assert exp_image.shape == shape
    assert exp_image.transform[:6] == crs_transform


@pytest.mark.parametrize(
    'base_image, crs, region, scale',
    [
        ('user_base_image', 'EPSG:3857', 'region_25ha', 0.1),
        ('user_fix_base_image', 'EPSG:32734', 'region_25ha', 30),
        ('l9_base_image', 'EPSG:3857', 'region_100ha', 10),
        ('s2_sr_hm_base_image', 'EPSG:32734', 'region_100ha', 30),
        ('modis_nbar_base_image', 'EPSG:3857', 'region_10000ha', 1000),
        ('google_dyn_world_base_image', 'EPSG:4326', 'region_10000ha', 1.0e-5),
    ],
)
def test_prepare_region_scale(base_image: str, crs: str, region: str, scale: float, request: pytest.FixtureRequest):
    """Test ``BaseImage._prepare_for_export()`` with ``region`` and ``scale`` parameters."""
    base_image: BaseImage = request.getfixturevalue(base_image)
    region: dict = request.getfixturevalue(region) if region else None

    exp_kwargs = dict(crs=crs, region=region, scale=scale)
    exp_image = base_image._prepare_for_export(**exp_kwargs)

    _test_export_image(exp_image, base_image, **exp_kwargs)


@pytest.mark.parametrize(
    'base_image, crs, region, shape',
    [
        ('user_base_image', 'EPSG:3857', 'region_25ha', (30, 40)),
        ('user_fix_base_image', 'EPSG:32734', 'region_100ha', (300, 400)),
        ('s2_sr_hm_base_image', 'EPSG:4326', 'region_10000ha', (300, 400)),
    ],
)
def test_prepare_region_shape(
    base_image: str, crs: str, region: str, shape: tuple[int, int], request: pytest.FixtureRequest
):
    """Test ``BaseImage._prepare_for_export()`` with ``region`` and ``shape`` parameters."""
    base_image: BaseImage = request.getfixturevalue(base_image)
    region: dict = request.getfixturevalue(region) if region else None

    exp_kwargs = dict(crs=crs, region=region, shape=shape)
    exp_image = base_image._prepare_for_export(**exp_kwargs)

    _test_export_image(exp_image, base_image, **exp_kwargs)


@pytest.mark.parametrize(
    'base_image, region',
    [
        ('user_fix_bnd_base_image', 'region_25ha'),
        ('l8_base_image', 'region_100ha'),
        ('l9_base_image', 'region_10000ha'),
        ('s2_sr_hm_base_image', 'region_25ha'),
        ('modis_nbar_base_image', 'region_100ha'),
        ('google_dyn_world_base_image', 'region_10000ha'),
    ],
)
def test_prepare_src_grid(base_image: str, region: str, request: pytest.FixtureRequest):
    """Test ``BaseImage._prepare_for_export()`` maintains the source pixel grid, with default value ``crs`` and
    ``scale`` arguments.
    """
    base_image: BaseImage = request.getfixturevalue(base_image)
    region: dict = request.getfixturevalue(region)

    exp_kwargs = dict(region=region)
    exp_image = base_image._prepare_for_export(**exp_kwargs)

    _test_export_image(exp_image, base_image, **exp_kwargs)


@pytest.mark.parametrize(
    'base_image, bands', [('s2_sr_hm_base_image', ['B1', 'B5']), ('l9_base_image', ['SR_B4', 'SR_B3', 'SR_B2'])]
)
def test_prepare_bands(base_image: str, bands: List[str], region_25ha: dict, request: pytest.FixtureRequest):
    """Test ``BaseImage._prepare_for_export()`` with ``bands`` parameter."""
    base_image: BaseImage = request.getfixturevalue(base_image)
    ref_image = BaseImage(base_image.ee_image.select(bands))
    exp_image = base_image._prepare_for_export(bands=bands)

    _test_export_image(exp_image, ref_image)


def test_prepare_bands_error(s2_sr_hm_base_image):
    """Test ``BaseImage._prepare_for_export()`` raises an error with incorrect bands."""
    with pytest.raises(ValueError) as ex:
        s2_sr_hm_base_image._prepare_for_export(bands=['unknown'])
    assert 'band' in str(ex.value)


def test_prepare_for_download(s2_sr_hm_base_image: BaseImage):
    """Test ``BaseImage._prepare_for_download()`` sets the rasterio profile as expected."""
    exp_image, exp_profile = s2_sr_hm_base_image._prepare_for_download()

    _test_export_image(exp_image, s2_sr_hm_base_image)

    # test dynamic values
    for key in ['dtype', 'count', 'crs', 'transform']:
        assert exp_profile[key] == getattr(exp_image, key), key

    assert exp_profile['width'] == exp_image.shape[1]
    assert exp_profile['height'] == exp_image.shape[0]
    assert exp_profile['nodata'] == _nodata_vals[exp_profile['dtype']]

    # test fixed values
    ref_profile = dict(driver='GTiff', compress='deflate', interleave='band', tiled=True, photometric='MINISBLACK')
    for key in ref_profile.keys():
        assert exp_profile[key] == ref_profile[key], key


@pytest.mark.parametrize('dtype, exp_nodata', _nodata_vals.items())
def test_prepare_nodata(user_fix_bnd_base_image: BaseImage, dtype: str, exp_nodata: float):
    """Test ``BaseImage._prepare_for_download()`` sets the ``nodata`` value correctly for different dtypes."""
    exp_image, exp_profile = user_fix_bnd_base_image._prepare_for_download(dtype=dtype)
    assert exp_image.dtype == dtype
    assert exp_profile['nodata'] == exp_nodata


@pytest.mark.parametrize(
    'src_image, dtype', [('s2_sr_hm_base_image', 'float32'), ('l9_base_image', None), ('modis_nbar_base_image', None)]
)
def test_scale_offset(src_image: str, dtype: str, region_100ha: Dict, request: pytest.FixtureRequest):
    """Test ``BaseImage._prepare_for_export(scale_offset=True)`` gives expected properties and reflectance ranges."""
    src_image: BaseImage = request.getfixturevalue(src_image)
    exp_image = src_image._prepare_for_export(scale_offset=True)

    assert exp_image.crs == src_image.crs
    assert exp_image.scale == src_image.scale
    assert exp_image.band_properties == src_image.band_properties
    assert exp_image.dtype == dtype or 'float64'

    def get_min_max_refl(base_image: BaseImage) -> tuple[dict, dict]:
        """Return the min & max of each reflectance band of ``base_image``."""
        band_props = base_image.band_properties
        refl_bands = [bp['name'] for bp in band_props if ('center_wavelength' in bp) and (bp['center_wavelength'] < 1)]
        ee_image = base_image.ee_image.select(refl_bands)
        min_max_dict = ee_image.reduceRegion(
            reducer=ee.Reducer.minMax(), geometry=region_100ha, bestEffort=True
        ).getInfo()
        min_dict = {k: v for k, v in min_max_dict.items() if 'min' in k}
        max_dict = {k: v for k, v in min_max_dict.items() if 'max' in k}
        return min_dict, max_dict

    # test the scaled and offset reflectance values lie between -0.5 and 1.5
    exp_min, exp_max = get_min_max_refl(exp_image)
    assert all(np.array(list(exp_min.values())) >= -0.5)
    assert all(np.array(list(exp_max.values())) <= 1.5)


def test_tile_shape():
    """Test ``BaseImage._get_tile_shape()`` satisfies the tile size limit for different image shapes."""
    max_tile_dim = 10000
    for max_tile_size, height, width in product(range(8, 32, 8), range(1, 11000, 1000), range(1, 11000, 1000)):
        exp_shape = (height, width)
        exp_image = BaseImageLike(shape=exp_shape)  # mock a BaseImage
        tile_shape, num_tiles = exp_image._get_tile_shape(max_tile_size=max_tile_size, max_tile_dim=max_tile_dim)
        tile_size = np.prod(tile_shape) * exp_image.count * np.dtype(exp_image.dtype).itemsize

        assert all(np.array(tile_shape) <= np.array(exp_shape))
        assert all(np.array(tile_shape) <= max_tile_dim)
        assert tile_size <= (max_tile_size << 20)


@pytest.mark.parametrize(
    'image_shape, tile_shape, image_transform',
    [
        ((1000, 500), (101, 101), Affine.identity()),
        ((1000, 100), (101, 101), Affine.scale(1.23)),
        ((1000, 102), (101, 101), Affine.scale(1.23) * Affine.translation(12, 34)),
    ],
)
def test_tiles(image_shape: tuple[int, int], tile_shape: tuple[int, int], image_transform: Affine):
    """Test continuity and coverage of tiles."""
    exp_image = BaseImageLike(shape=image_shape, transform=image_transform)
    tiles = [tile for tile in exp_image._tiles(tile_shape=tile_shape)]

    # test tile continuity
    prev_tile = tiles[0]
    accum_window = prev_tile.window
    for tile in tiles[1:]:
        accum_window = windows.union(accum_window, tile.window)
        assert all(np.array(tile._shape) <= np.array(tile_shape))

        if tile.window.row_off == prev_tile.window.row_off:
            assert tile.window.col_off == (prev_tile.window.col_off + prev_tile.window.width)
            ref_transform = prev_tile._transform * Affine.translation(prev_tile.window.width, 0)
        else:
            assert tile.window.row_off == (prev_tile.window.row_off + prev_tile.window.height)
            ref_transform = prev_tile._transform * Affine.translation(
                -prev_tile.window.col_off, prev_tile.window.height
            )

        assert tile._transform == pytest.approx(ref_transform, abs=1e-9)
        prev_tile = tile

    # test exp_image is fully covered by tiles
    assert (accum_window.height, accum_window.width) == exp_image.shape


def test_download_transform_shape(user_base_image: BaseImage, tmp_path: pathlib.Path):
    """Test ``BaseImage.download()`` file properties and contents with ``crs_transform`` and ``shape`` arguments."""
    # reference profile to test against
    ref_profile = dict(
        crs='EPSG:3857', transform=Affine(1, 0, 0, 0, -1, 0), width=10, height=10, dtype='uint16', count=3
    )

    # form export kwargs from ref_profile
    shape = (ref_profile['height'], ref_profile['width'])
    exp_kwargs = dict(
        crs=ref_profile['crs'], crs_transform=ref_profile['transform'], shape=shape, dtype=ref_profile['dtype']
    )

    # download
    filename = tmp_path.joinpath('test.tif')
    user_base_image.download(filename, **exp_kwargs)
    assert filename.exists()

    # test file format and contents
    with rio.open(filename, 'r') as ds:
        for key in ref_profile:
            assert ds.profile[key] == ref_profile[key]

        array = ds.read()
        for i in range(ds.count):
            assert np.all(array[i] == i + 1)


def test_download_region_scale(user_base_image: BaseImage, tmp_path: pathlib.Path):
    """Test ``BaseImage.download()`` file properties and contents with ``region`` and ``scale`` arguments."""
    # reference profile to test against
    ref_profile = dict(
        crs='EPSG:3857', transform=Affine(1, 0, 0, 0, -1, 0), width=10, height=10, dtype='uint16', count=3
    )

    # form export kwargs from ref_profile
    shape = (ref_profile['height'], ref_profile['width'])
    bounds = windows.bounds(windows.Window(0, 0, *shape[::-1]), ref_profile['transform'])
    region = bounds_polygon(*bounds, crs=ref_profile['crs'])
    exp_kwargs = dict(
        crs=ref_profile['crs'], region=region, scale=ref_profile['transform'][0], dtype=ref_profile['dtype']
    )

    # download
    filename = tmp_path.joinpath('test.tif')
    user_base_image.download(filename, **exp_kwargs)
    assert filename.exists()

    # test file format and contents
    with rio.open(filename, 'r') as ds:
        for key in ref_profile:
            assert ds.profile[key] == ref_profile[key]
        array = ds.read()
        for i in range(ds.count):
            assert np.all(array[i] == i + 1)


def test_overviews(user_base_image: BaseImage, region_25ha: Dict, tmp_path: pathlib.Path):
    """Test overviews get built by ``BaseImage.download()``."""
    filename = tmp_path.joinpath('test_user_download.tif')
    user_base_image.download(filename, region=region_25ha, crs='EPSG:3857', scale=1)
    assert filename.exists()
    with rio.open(filename, 'r') as ds:
        for bi in ds.indexes:
            assert len(ds.overviews(bi)) > 0
            assert ds.overviews(bi)[0] == 2


def test_metadata(s2_sr_hm_base_image: BaseImage, region_25ha: Dict, tmp_path: pathlib.Path):
    """Test metadata is written by ``BaseImage.download()``."""
    filename = tmp_path.joinpath('test_s2_band_subset_download.tif')
    s2_sr_hm_base_image.download(filename, region=region_25ha, crs='EPSG:3857', scale=60, bands=['B9'])
    assert filename.exists()
    with rio.open(filename, 'r') as ds:
        assert 'LICENSE' in ds.tags()
        assert len(ds.tags()['LICENSE']) > 0
        assert 'B9' in ds.descriptions
        band_dict = ds.tags(1)
        for key in ['gsd', 'name', 'description']:
            assert key in band_dict


@pytest.mark.parametrize('type', ExportType)
def test_start_export(type: ExportType, user_fix_base_image: BaseImage, region_25ha: Dict):
    """Test ``BaseImage.export()`` starts a small export."""
    # Note that this export should start successfully, but will ultimately fail for the asset and cloud options.  For
    # the asset option, there will be an overwrite issue.  For cloud storage, there is no 'geedim' bucket.
    task = user_fix_base_image.export(
        'test_export', type=type, folder='geedim', scale=30, region=region_25ha, wait=False
    )
    assert task.active()
    assert task.status()['state'] == 'READY' or task.status()['state'] == 'RUNNING'


def __test_export_asset(user_fix_base_image: BaseImage, region_25ha: Dict):
    """Test start of a small export to EE asset."""
    # TODO: consider removing this slow test in favour of the equivalent integration test
    filename = f'test_export_{np.random.randint(1<<31)}'
    folder = 'geedim'
    asset_id = utils.asset_id(filename, folder)
    # Note: to allow parallel tests exporting to assets, we use random asset names to prevent conflicts.  The test
    # must wait for export to complete and clean up after itself, otherwise lots of test assets will accumulate in my
    # geedim cloud project.

    try:
        # export and test asset exists
        task = user_fix_base_image.export(
            filename, type=ExportType.asset, folder=folder, scale=30, region=region_25ha, wait=True
        )
        assert task.status()['state'] == 'COMPLETED'
        assert ee.data.getAsset(asset_id) is not None
    finally:
        # delete asset
        try:
            ee.data.deleteAsset(asset_id)
        except ee.ee_exception.EEException:
            pass


def test_download_bigtiff(s2_sr_hm_base_image: BaseImage, tmp_path: pathlib.Path):
    """Test ``BaseImage._prepare_for_download()`` sets the ``bigtiff`` profile item for images larger than 4GB."""
    exp_image, profile = s2_sr_hm_base_image._prepare_for_download()
    assert exp_image.size >= 4e9
    assert 'bigtiff' in profile
    assert profile['bigtiff']


def test_prepare_ee_geom(l9_base_image: BaseImage, tmp_path: pathlib.Path):
    """Test that ``BaseImage._prepare_for_export()`` works with an ee.Geometry region at native CRS and scale (Issue
    #6).
    """
    region = l9_base_image.ee_image.geometry()
    exp_image = l9_base_image._prepare_for_export(region=region)
    assert exp_image.scale == l9_base_image.scale


@pytest.mark.parametrize(
    'base_image, exp_value',
    [
        ('s2_sr_hm_base_image', True),
        ('l9_base_image', True),
        ('modis_nbar_base_image_unbnd', False),
        ('user_base_image', False),
        ('user_fix_base_image', False),
    ],
)
def test_bounded(base_image: str, exp_value: bool, request: pytest.FixtureRequest):
    """Test ``BaseImage.bounded`` for different images."""
    base_image: BaseImage = request.getfixturevalue(base_image)
    assert base_image.bounded == exp_value


# TODO:
#  - test float mask/nodata in downloaded image



"""
   Copyright 2021 Dugal Harris - dugalh@gmail.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import logging
import threading
import time
from io import BytesIO

import numpy as np
import rasterio as rio
import requests
from rasterio import Affine
from rasterio.errors import RasterioIOError
from rasterio.windows import Window
from requests.exceptions import JSONDecodeError, RequestException
from tqdm.auto import tqdm

from geedim import utils
from geedim.errors import TileError

logger = logging.getLogger(__name__)


class Tile:
    # lock to prevent concurrent calls to ee.Image.getDownloadURL(), which can cause a seg fault in the standard
    # python networking libraries.
    _ee_lock = threading.Lock()

    def __init__(self, exp_image, window: Window):
        """
        Class for downloading an Earth Engine image tile (a rectangular region of interest in the image).

        Parameters
        ----------
        exp_image: BaseImage
            BaseImage instance to derive the tile from.
        window: Window
            rasterio window into `exp_image`, specifying the region of interest for this tile.
        """
        self._exp_image = exp_image
        self._window = window
        # offset the image geo-transform origin so that it corresponds to the UL corner of the tile.
        self._transform = exp_image.transform * Affine.translation(window.col_off, window.row_off)
        self._shape = (window.height, window.width)

    @property
    def window(self) -> Window:
        """rasterio tile window into the source image."""
        return self._window

    @staticmethod
    def _raise_for_status(response: requests.Response):
        """Raise a TileError if the tile cannot be downloaded."""
        download_size = int(response.headers.get('content-length', 0))
        if download_size == 0 or not response.status_code == 200:
            msg = f'Error downloading tile: {response.status_code} - {response.reason}. URL: {response.url}.'
            try:
                resp_dict = response.json()
                if 'error' in resp_dict and 'message' in resp_dict['error']:
                    # raise an exception with the response error message
                    msg = resp_dict['error']['message']
                    msg = f'Error downloading tile: {msg} URL: {response.url}.'
                    if 'user memory limit exceeded' in msg.lower():
                        msg += (
                            '\nThe `max_tile_size` or `max_tile_dim` parameters can be decreased to work around this '
                            'error.  Alternatively you can export to Earth Engine asset, and then download the asset '
                            'image.'
                        )
            except JSONDecodeError:
                pass

            raise TileError(msg)

    def _download_to_array(self, url: str, session: requests.Session = None, bar: tqdm = None) -> np.ndarray:
        """Download the image tile into a numpy array."""
        # get image download response
        session = session or requests
        response = session.get(url, stream=True, timeout=(30, 300))

        # raise a TileError if the tile cannot be downloaded
        self._raise_for_status(response)

        # find raw download size
        download_size = int(response.headers.get('content-length', 0))
        dtype_size = np.dtype(self._exp_image.dtype).itemsize
        raw_download_size = self._shape[0] * self._shape[1] * self._exp_image.count * dtype_size

        # download & read the tile
        downloaded_size = 0
        buf = BytesIO()
        try:
            # download gtiff into buffer
            for data in response.iter_content(chunk_size=10240):
                if bar:
                    # update with raw download progress (0-1)
                    bar.update(raw_download_size * (len(data) / download_size))
                buf.write(data)
                downloaded_size += len(data)
            buf.flush()

            # read the tile array from the GeoTIFF buffer
            buf.seek(0)
            env = rio.Env(GDAL_NUM_THREADS='ALL_CPUs', GTIFF_FORCE_RGBA=False)
            with utils.suppress_rio_logs(), env, rio.open(buf, 'r') as ds:
                array = ds.read()
            return array

        except (RequestException, RasterioIOError):
            if bar:
                # reverse progress bar
                bar.update(-raw_download_size * (downloaded_size / download_size))
                pass
            raise

    def download(
        self,
        session: requests.Session = None,
        bar: tqdm = None,
        max_retries: int = 5,
        backoff_factor: float = 2.0,
    ) -> np.ndarray:
        """
        Download the image tile into a numpy array.

        Parameters
        ----------
        session: requests.Session, optional
            requests session to use for downloading
        bar: tqdm, optional
            tqdm progress bar instance to update with incremental (0-1) download progress.
        max_retries: int, optional
            Number of times to retry downloading the tile.  This is independent of the ``session``, which may have its
            own retry configuration.
        backoff_factor: float, optional
            Backoff factor to apply between tile download retries.  The delay between retries is: {backoff_factor} *
            (2 ** ({number of previous retries})) seconds.  This is independent of the ``session``, which may have its
            own retry configuration.

        Returns
        -------
        array: numpy.ndarray
            3D numpy array of the tile pixel data with bands down the first dimension.
        """
        session = session or requests

        # get download URL
        url = self._exp_image.ee_image.getDownloadURL(
            dict(
                crs=self._exp_image.crs,
                crs_transform=tuple(self._transform)[:6],
                dimensions=self._shape[::-1],
                format='GEO_TIFF',
            )
        )

        # download and read the tile, with retries
        for retry in range(max_retries + 1):
            try:
                return self._download_to_array(url, session=session, bar=bar)
            except (RequestException, RasterioIOError) as ex:
                if retry < max_retries:
                    time.sleep(backoff_factor * (2**retry))
                    logger.warning(f'Tile download failed, retry {retry + 1} of {max_retries}.  URL: {url}. {str(ex)}.')
                else:
                    raise TileError(f'Tile download failed, reached the maximum retries.  URL: {url}.') from ex



"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

from datetime import datetime, timezone
from typing import Dict, List, Union

import ee
import numpy as np
import pytest

from geedim import schema
from geedim.collection import MaskedCollection
from geedim.enums import CompositeMethod, ResamplingMethod
from geedim.errors import InputImageError, UnfilteredError
from geedim.mask import MaskedImage
from geedim.utils import get_projection, split_id


@pytest.fixture()
def l4_5_images(l4_image_id, l5_masked_image) -> List[Union[str, MaskedImage]]:
    """A list of landsat 4 & 5 image ID / MaskedImage's"""
    return [l4_image_id, l5_masked_image]


@pytest.fixture()
def l8_9_images(l8_image_id, l9_masked_image) -> List[Union[str, MaskedImage]]:
    """A list of landsat 8 & 9 image IDs/ MaskedImage's"""
    return [l8_image_id, l9_masked_image]


@pytest.fixture()
def s2_sr_hm_images(s2_sr_hm_image_ids: list[str]) -> list[str | MaskedImage]:
    """A list of harmonised Sentinel-2 SR image IDs / MaskedImage's with QA* data, covering `region_*ha` with partial
    cloud/shadow.
    """
    image_list = s2_sr_hm_image_ids.copy()
    image_list[-1] = MaskedImage.from_id(image_list[-1])
    return image_list


@pytest.fixture()
def gedi_image_list() -> List[Union[str, MaskedImage]]:
    """A list of GEDI canopy top height IDs/ MaskedImage's"""
    return [
        'LARSE/GEDI/GEDI02_A_002_MONTHLY/202008_018E_036S',
        'LARSE/GEDI/GEDI02_A_002_MONTHLY/202009_018E_036S',
        MaskedImage.from_id('LARSE/GEDI/GEDI02_A_002_MONTHLY/202005_018E_036S'),
    ]


def test_split_id():
    """Test split_id()."""
    coll_name, im_id = split_id('A/B/C')
    assert coll_name == 'A/B'
    assert im_id == 'C'
    coll_name, im_id = split_id('ABC')
    assert coll_name == ''
    assert im_id == 'ABC'


@pytest.mark.parametrize('name', ['s2_sr_hm_image_id', 'l9_image_id', 'gch_image_id'])
def test_from_name(name: str, request):
    """Test MaskedCollection.from_name()."""
    name = request.getfixturevalue(name)
    name, _ = split_id(name)
    gd_collection = MaskedCollection.from_name(name)
    assert gd_collection._name == name
    assert gd_collection.schema is not None
    assert len(gd_collection.schema) >= len(schema.default_prop_schema)
    assert gd_collection.ee_collection == ee.ImageCollection(name)


def test_unfiltered_error(s2_sr_hm_image_id):
    """Test UnfilteredError is raised when calling `properties` or `composite` on an unfiltered collection."""
    gd_collection = MaskedCollection.from_name(split_id(s2_sr_hm_image_id)[0])
    with pytest.raises(UnfilteredError):
        _ = gd_collection.properties
    with pytest.raises(UnfilteredError):
        _ = gd_collection.composite()


def test_from_list_errors(landsat_image_ids, user_masked_image):
    """Test MaskedCollection.from_list() various error cases."""
    with pytest.raises(InputImageError):
        # test an error is raised when an image has no 'id'/'system:time_start' property
        MaskedCollection.from_list([landsat_image_ids[0], user_masked_image])

    with pytest.raises(InputImageError):
        # test an error is raised when images are not from compatible collections
        MaskedCollection.from_list([*landsat_image_ids])

    with pytest.raises(TypeError):
        # test an error is raised when an image type is not recognised
        MaskedCollection.from_list([landsat_image_ids[0], {}])

    with pytest.raises(ValueError):
        # test an error is raised when image list is empty
        MaskedCollection.from_list([])


@pytest.mark.parametrize('image_list', ['l4_5_images', 'l8_9_images'])
def test_from_list_landsat(image_list: str, request):
    """
    Test MaskedCollection.from_list() works with landsat images from different, but spectrally compatible
    collections.
    """
    image_list: List = request.getfixturevalue(image_list)
    gd_collection = MaskedCollection.from_list(image_list)
    assert 'LANDSAT' in gd_collection.name
    assert gd_collection.properties is not None


@pytest.mark.parametrize('image_list', ['s2_sr_hm_images', 'gedi_image_list'])
def test_from_list(image_list: str, request):
    """
    Test MaskedCollection.from_list() generates a valid MaskedCollection object from lists of cloud/shadow
    maskable, and generic images.
    """
    image_list: List = request.getfixturevalue(image_list)
    image_ids = [im_obj if isinstance(im_obj, str) else im_obj.id for im_obj in image_list]
    gd_collection = MaskedCollection.from_list(image_list)
    assert gd_collection.properties is not None
    assert gd_collection.schema is not None
    assert len(gd_collection.schema) >= len(schema.default_prop_schema)
    assert gd_collection.name == split_id(image_ids[0])[0]
    assert len(gd_collection.properties) == len(image_list)
    assert list(gd_collection.properties.keys()) == image_ids
    assert set(gd_collection.schema.keys()) > set(list(gd_collection.properties.values())[0].keys())
    assert gd_collection.properties_table is not None
    assert gd_collection.schema_table is not None


@pytest.mark.parametrize('image_list', ['s2_sr_hm_images', 'gedi_image_list'])
def test_from_list_order(image_list: str, request):
    """Test MaskedCollection.from_list() maintains the order of the provided image list."""
    image_list: List = request.getfixturevalue(image_list)[::-1]
    image_ids = [im_obj if isinstance(im_obj, str) else im_obj.id for im_obj in image_list]
    gd_collection = MaskedCollection.from_list(image_list)
    assert list(gd_collection.properties.keys()) == image_ids


def test_from_list_ee_image(gedi_image_list: List):
    """Test MaskedCollection.from_list() with an ee.Image in the list."""
    image_ids = [im_obj if isinstance(im_obj, str) else im_obj.id for im_obj in gedi_image_list]
    image_list = gedi_image_list
    image_list[1] = ee.Image(image_list[1])
    gd_collection = MaskedCollection.from_list(image_list)
    assert list(gd_collection.properties.keys()) == image_ids


@pytest.mark.parametrize(
    'image_list, add_props',
    [
        ('s2_sr_hm_images', ['AOT_RETRIEVAL_ACCURACY', 'CLOUDY_PIXEL_PERCENTAGE']),
        ('l8_9_images', ['CLOUD_COVER', 'GEOMETRIC_RMSE_VERIFY']),
    ],
)
def test_from_list_add_props(image_list: str, add_props: List, request: pytest.FixtureRequest):
    """
    Test MaskedCollection.from_list(add_props=...) contains the add_props in properties and schema.
    """
    image_list: List = request.getfixturevalue(image_list)
    gd_collection = MaskedCollection.from_list(image_list, add_props=add_props)
    assert gd_collection.properties is not None
    assert gd_collection.schema is not None
    assert len(gd_collection.schema) > len(schema.default_prop_schema)
    assert all([add_prop in gd_collection.schema.keys() for add_prop in add_props])
    assert all([gd_collection.schema[add_prop]['abbrev'] is not None for add_prop in add_props])
    assert all([add_prop in list(gd_collection.properties.values())[0].keys() for add_prop in add_props])
    assert gd_collection.properties_table is not None
    assert gd_collection.schema_table is not None


@pytest.mark.parametrize(
    'name, start_date, end_date, region, fill_portion, cloudless_portion, is_csmask',
    [
        ('LANDSAT/LC09/C02/T1_L2', '2022-01-01', '2022-02-01', 'region_100ha', 0, 50, True),
        ('LANDSAT/LE07/C02/T1_L2', '2022-01-01', '2022-02-01', 'region_100ha', 0, 0, True),
        ('LANDSAT/LT05/C02/T1_L2', '2005-01-01', '2006-02-01', 'region_100ha', 40, 50, True),
        ('COPERNICUS/S2_SR', '2022-01-01', '2022-01-15', 'region_100ha', 0, 50, True),
        ('COPERNICUS/S2_HARMONIZED', '2022-01-01', '2022-01-15', 'region_100ha', 50, 40, True),
        ('LARSE/GEDI/GEDI02_A_002_MONTHLY', '2021-08-01', '2021-09-01', 'region_100ha', 0.1, 0, False),
    ],
)
def test_search(
    name, start_date: str, end_date: str, region: str, fill_portion: float, cloudless_portion: float, is_csmask, request
):
    """
    Test MaskedCollection.search() with fill / cloudless portion filters gives valid results for different cloud/shadow
    maskable, and generic collections.
    """
    region: Dict = request.getfixturevalue(region)
    gd_collection = MaskedCollection.from_name(name)
    searched_collection = gd_collection.search(
        start_date, end_date, region, fill_portion=fill_portion, cloudless_portion=cloudless_portion
    )

    properties = searched_collection.properties
    start_date = datetime.strptime(start_date, '%Y-%m-%d')
    end_date = datetime.strptime(end_date, '%Y-%m-%d')
    im_dates = np.array(
        [datetime.fromtimestamp(im_props['system:time_start'] / 1000) for im_props in properties.values()]
    )
    # test FILL_PORTION in expected range
    im_fill_portions = np.array([im_props['FILL_PORTION'] for im_props in properties.values()])
    assert np.all(im_fill_portions >= fill_portion) and np.all(im_fill_portions <= 100)
    if is_csmask:  # is a cloud/shadow masked collection
        # test CLOUDLESS_PORTION in expected range
        im_cl_portions = np.array([im_props['CLOUDLESS_PORTION'] for im_props in properties.values()])
        assert np.all(im_cl_portions >= cloudless_portion) and np.all(im_cl_portions <= 100)
    # test search result image dates lie between `start_date` and `end_date`
    assert np.all(im_dates >= start_date) and np.all(im_dates < end_date)
    assert set(searched_collection.schema.keys()) == set(list(properties.values())[0].keys())
    # test search result image dates are sorted
    assert np.all(sorted(im_dates) == im_dates)


def test_empty_search(region_100ha):
    """Test MaskedCollection.search() for empty search results."""
    gd_collection = MaskedCollection.from_name('LANDSAT/LC09/C02/T1_L2')
    searched_collection = gd_collection.search('2022-01-01', '2022-01-02', region_100ha, cloudless_portion=100)
    assert searched_collection.properties is not None
    assert len(searched_collection.properties) == 0
    assert searched_collection.properties_table is not None


def test_search_no_end_date(region_100ha):
    """Test MaskedCollection.search() with ``end_date=None`` searches for a day from ``start_date``."""
    gd_collection = MaskedCollection.from_name('LANDSAT/LC09/C02/T1_L2')
    searched_collection = gd_collection.search('2022-01-03', None, region_100ha)

    start_date = datetime.strptime('2022-01-03', '%Y-%m-%d')
    end_date = datetime.strptime('2022-01-04', '%Y-%m-%d')
    properties = searched_collection.properties
    im_dates = np.array(
        [datetime.fromtimestamp(im_props['system:time_start'] / 1000) for im_props in properties.values()]
    )
    assert len(properties) > 0
    assert np.all(im_dates >= start_date) and np.all(im_dates < end_date)


def test_search_date_error(region_100ha):
    """Test MaskedCollection.search() raises an error when end date is on or before start date."""
    gd_collection = MaskedCollection.from_name('LANDSAT/LC09/C02/T1_L2')
    with pytest.raises(ValueError):
        _ = gd_collection.search('2022-01-02', '2022-01-01', region_100ha)


def test_search_mult_kwargs(region_100ha):
    """
    When a search filtered collection is searched again, test that masks change with different cloud/shadow kwargs.
    """
    start_date = '2022-01-01'
    end_date = '2022-01-10'
    gd_collection = MaskedCollection.from_name('COPERNICUS/S2_SR_HARMONIZED')

    def get_cloudless_portion(properties: Dict) -> List[float]:
        return [prop_dict['CLOUDLESS_PORTION'] for prop_dict in properties.values()]

    filt_collection = gd_collection.search(
        start_date, end_date, region_100ha, mask_method='cloud-score', score=0.5, fill_portion=0
    )
    filt_coll_prob80 = filt_collection.search(
        start_date, end_date, region_100ha, mask_method='cloud-score', score=0.5, fill_portion=0
    )
    filt_coll_prob40 = filt_collection.search(
        start_date, end_date, region_100ha, mask_method='cloud-score', score=0.2, fill_portion=0
    )

    cp_ref = get_cloudless_portion(filt_collection.properties)
    cp_prob80 = get_cloudless_portion(filt_coll_prob80.properties)
    cp_prob40 = get_cloudless_portion(filt_coll_prob40.properties)

    assert cp_ref == pytest.approx(cp_prob80, abs=1e-3)
    assert cp_ref != pytest.approx(cp_prob40, abs=1e-1)


def test_search_custom_filter(region_25ha):
    """
    Test that a CLOUDLESS_PORTION custom filter gives the same search results as the equivalent cloudless_portion
    kwarg specification.
    """
    start_date = '2022-01-01'
    end_date = '2022-02-01'
    gd_collection = MaskedCollection.from_name('LANDSAT/LC09/C02/T1_L2')
    kwarg_coll = gd_collection.search(start_date, end_date, region_25ha, cloudless_portion=90)
    cust_filt_coll = gd_collection.search(start_date, end_date, region_25ha, custom_filter='CLOUDLESS_PORTION>=90')
    assert (kwarg_coll.properties is not None) and (len(kwarg_coll.properties) > 0)
    assert kwarg_coll.properties == cust_filt_coll.properties


def test_search_add_props(region_25ha):
    """
    Test that specified add_props are added to the search results.
    """
    add_props = ['CLOUD_COVER', 'GEOMETRIC_RMSE_VERIFY']
    gd_collection = MaskedCollection.from_name('LANDSAT/LC09/C02/T1_L2', add_props=add_props)
    searched_coll = gd_collection.search('2022-01-01', '2022-02-01', region_25ha)
    assert all([add_prop in searched_coll.schema.keys() for add_prop in add_props])
    assert all([add_prop in list(searched_coll.properties.values())[0].keys() for add_prop in add_props])
    assert searched_coll.properties_table is not None
    assert searched_coll.schema_table is not None


@pytest.mark.parametrize(
    'name, start_date, end_date, region',
    [
        ('LANDSAT/LC09/C02/T1_L2', '2022-01-01', '2022-02-01', 'region_100ha'),
        ('COPERNICUS/S2_HARMONIZED', '2022-01-01', '2022-01-15', 'region_100ha'),
        ('LARSE/GEDI/GEDI02_A_002_MONTHLY', '2021-08-01', '2021-09-01', 'region_100ha'),
    ],
)
def test_search_no_fill_or_cloudless_portion(
    name: str, start_date: str, end_date: str, region: str, request: pytest.FixtureRequest
):
    """
    Test MaskedCollection.search() without fill / cloudless portion filters gives valid results for different
    collections.
    """
    region: Dict = request.getfixturevalue(region)
    gd_collection = MaskedCollection.from_name(name)
    searched_collection = gd_collection.search(start_date, end_date, region)

    properties = searched_collection.properties
    start_date = datetime.strptime(start_date, '%Y-%m-%d')
    end_date = datetime.strptime(end_date, '%Y-%m-%d')
    im_dates = np.array(
        [datetime.fromtimestamp(im_props['system:time_start'] / 1000) for im_props in properties.values()]
    )
    # test FILL_PORTION and CLOUDLESS_PORTION are not in properties
    prop_keys = list(properties.values())[0].keys()
    assert 'FILL_PORTION' not in prop_keys
    assert 'CLOUDLESS_PORTION' not in prop_keys
    # test search result image dates lie between `start_date` and `end_date`
    assert np.all(im_dates >= start_date) and np.all(im_dates < end_date)
    assert set(searched_collection.schema.keys()) >= set(list(properties.values())[0].keys())
    # test search result image dates are sorted
    assert np.all(sorted(im_dates) == im_dates)


@pytest.mark.parametrize(
    'image_list, method, region, date',
    [
        ('s2_sr_hm_images', CompositeMethod.q_mosaic, 'region_10000ha', None),
        ('s2_sr_hm_images', CompositeMethod.q_mosaic, None, '2021-10-01'),
        ('gedi_image_list', CompositeMethod.mosaic, 'region_10000ha', None),
        ('gedi_image_list', CompositeMethod.mosaic, None, '2020-09-01'),
        ('l8_9_images', CompositeMethod.medoid, 'region_10000ha', None),
        ('l8_9_images', CompositeMethod.medoid, None, '2021-10-01'),
    ],
)
def test_composite_region_date_ordering(image_list, method, region, date, request):
    """
    In MaskedCollection.composite(), test the component images are ordered correctly, according to `region`/`date`
    parameters.
    """
    image_list: List = request.getfixturevalue(image_list)
    region: Dict = request.getfixturevalue(region) if region else None
    gd_collection = MaskedCollection.from_list(image_list)
    ee_collection = gd_collection._prepare_for_composite(method=method, date=date, region=region)
    properties = gd_collection._get_properties(ee_collection)
    assert len(properties) == len(image_list)

    if region:
        # test images are ordered by CLOUDLESS/FILL_PORTION, and that portions are not the same
        schema_keys = list(gd_collection.schema.keys())
        portion_key = 'CLOUDLESS_PORTION' if 'CLOUDLESS_PORTION' in schema_keys else 'FILL_PORTION'
        im_portions = [im_props[portion_key] for im_props in properties.values()]
        assert sorted(im_portions) == im_portions
        assert len(set(im_portions)) == len(im_portions)

    elif date:
        # test images are ordered by time difference with `date`
        im_dates = np.array(
            [datetime.fromtimestamp(im_props['system:time_start'] / 1000) for im_props in properties.values()]
        )
        comp_date = datetime.strptime(date, '%Y-%m-%d')
        im_date_diff = np.abs(comp_date - im_dates)
        assert all(sorted(im_date_diff, reverse=True) == im_date_diff)


def _get_masked_portion(ee_image: ee.Image, proj: ee.Projection = None, region: dict = None) -> ee.Number:
    """Return the valid portion of the ``ee_image`` inside ``region``.  Assumes the ``region`` is completely covered
    by ``ee_image``.
    """
    proj = proj or get_projection(ee_image)
    ee_mask = ee_image.select('SR_B.*|B.*|rh.*').mask().reduce(ee.Reducer.allNonZero()).rename('EE_MASK')
    mean = ee_mask.reduceRegion(reducer='mean', crs=proj, scale=proj.nominalScale(), geometry=region)
    return ee.Number(mean.get('EE_MASK')).multiply(100)


@pytest.mark.parametrize(
    'image_list, method, mask',
    [
        ('s2_sr_hm_images', CompositeMethod.q_mosaic, True),
        ('s2_sr_hm_images', CompositeMethod.mosaic, False),
        ('l8_9_images', CompositeMethod.medoid, True),
        ('l8_9_images', CompositeMethod.median, False),
        ('s2_sr_hm_images', CompositeMethod.medoid, True),
        ('s2_sr_hm_images', CompositeMethod.medoid, False),
        ('l8_9_images', CompositeMethod.medoid, False),
    ],
)
def test_composite_mask(image_list, method, mask, region_100ha, request):
    """In MaskedImage.composite(), test masking of component and composite images with the `mask` parameter."""
    # TODO: combine >1 getInfo() calls into 1 for all tests
    # form the composite collection and image
    image_list: List = request.getfixturevalue(image_list)
    gd_collection = MaskedCollection.from_list(image_list)
    ee_collection = gd_collection._prepare_for_composite(method=method, mask=mask)
    composite_im = gd_collection.composite(method=method, mask=mask)
    properties = gd_collection._get_properties(ee_collection)
    proj = get_projection(ee_collection.first(), min_scale=True)
    assert len(properties) == len(image_list)

    def get_masked_portions(ee_image: ee.Image, portions: ee.List) -> ee.List:
        """Append the valid portion of ``ee_image`` to  ``portions``."""
        portion = _get_masked_portion(ee_image, proj=proj, region=region_100ha)
        return ee.List(portions).add(portion)

    # get the mask portions for the component and composite images
    component_portions = ee_collection.iterate(get_masked_portions, ee.List([]))
    composite_portion = _get_masked_portion(composite_im.ee_image, proj=proj, region=region_100ha)
    component_portions, composite_portion = ee.List([component_portions, composite_portion]).getInfo()
    component_portions = np.array(component_portions)

    # test masking of components and composite image
    if mask:
        assert np.all(component_portions > 0) and np.all(component_portions < 100)
        assert composite_portion <= np.sum(component_portions)
        assert composite_portion >= np.min(component_portions)
    else:
        assert component_portions == pytest.approx(100, abs=2)
        assert composite_portion >= component_portions.max()


@pytest.mark.parametrize(
    'masked_image, method, mask_kwargs',
    [
        ('s2_sr_hm_nocp_masked_image', 'q-mosaic', dict(mask_method='cloud-prob')),
        ('s2_sr_hm_qa_zero_masked_image', 'medoid', dict(mask_method='qa')),
        ('s2_sr_hm_nocs_masked_image', 'median', dict(mask_method='cloud-score')),
    ],
)
def test_s2_composite_mask_missing_data(masked_image: str, method: str, mask_kwargs: dict, region_100ha, request):
    """In MaskedImage.composite(), test when an S2 component image is fully masked due to missing cloud data,
    the composite image is also fully masked.
    """
    # form the composite collection and image
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    image_list = [masked_image]
    gd_collection = MaskedCollection.from_list(image_list)

    ee_collection = gd_collection._prepare_for_composite(method=method, mask=True, **mask_kwargs)
    composite_im = gd_collection.composite(method=method, mask=True, **mask_kwargs)
    properties = gd_collection._get_properties(ee_collection)
    proj = get_projection(ee_collection.first(), min_scale=True)
    assert len(properties) == len(image_list)

    # get the mask portions for the component and composite images
    component_portion = _get_masked_portion(ee_collection.first(), proj=proj, region=region_100ha).getInfo()
    composite_portion = _get_masked_portion(composite_im.ee_image, proj=proj, region=region_100ha).getInfo()

    # test component and composite images are fully masked
    assert component_portion == composite_portion == 0


def test_s2_composite_q_mosaic_missing_data(s2_sr_hm_nocs_masked_image: MaskedImage, region_100ha: dict):
    """In MaskedImage.composite(), test when an S2 component image is unmasked, but has masked CLOUD_DIST band due to
    missing cloud data, the composite image is fully masked with 'q-mosaic' method.
    """
    # form the composite collection and image
    image_list = [s2_sr_hm_nocs_masked_image]
    gd_collection = MaskedCollection.from_list(image_list)

    kwargs = dict(method='q-mosaic', mask_method='cloud-score', mask=False)
    ee_collection = gd_collection._prepare_for_composite(**kwargs)
    composite_im = gd_collection.composite(**kwargs)
    properties = gd_collection._get_properties(ee_collection)
    proj = get_projection(ee_collection.first(), min_scale=True)
    assert len(properties) == len(image_list)

    # get and test the mask portions for the component and composite images
    component_portion = _get_masked_portion(ee_collection.first(), proj=proj, region=region_100ha).getInfo()
    composite_portion = _get_masked_portion(composite_im.ee_image, proj=proj, region=region_100ha).getInfo()
    assert component_portion == pytest.approx(100, abs=1)
    assert composite_portion == 0


@pytest.mark.parametrize(
    'image_list, resampling, scale',
    [
        ('s2_sr_hm_images', ResamplingMethod.bilinear, 7.5),
        ('s2_sr_hm_images', ResamplingMethod.bicubic, 7.5),
        ('s2_sr_hm_images', ResamplingMethod.average, 30),
        ('l8_9_images', ResamplingMethod.bilinear, 20),
        ('l8_9_images', ResamplingMethod.bicubic, 20),
        ('l8_9_images', ResamplingMethod.average, 50),
    ],
)
def test_composite_resampling(
    image_list: str,
    resampling: ResamplingMethod,
    scale: float,
    region_100ha: Dict,
    request: pytest.FixtureRequest,
):
    """Test that resampling smooths the composite image."""
    image_list: List = request.getfixturevalue(image_list)
    gd_collection = MaskedCollection.from_list(image_list)
    comp_im = gd_collection.composite(method=CompositeMethod.mosaic, mask=False)
    comp_im_resampled = gd_collection.composite(method=CompositeMethod.mosaic, resampling=resampling, mask=False)

    # find mean of std deviations of reflectance bands for each composite image
    crs = gd_collection.ee_collection.first().select(0).projection().crs()
    stds = []
    for im in [comp_im, comp_im_resampled]:
        im = im.ee_image.select(gd_collection.refl_bands)
        im = im.reproject(crs=crs, scale=scale)  # required to resample at scale
        std = im.reduceRegion('stdDev', geometry=region_100ha).values().reduce('mean')
        stds.append(std)
    stds = ee.List(stds).getInfo()

    # test comp_im_resampled is smoother than comp_im
    assert stds[1] < stds[0]


def test_composite_s2_cloud_mask_params(s2_sr_hm_images, region_100ha):
    """
    Test cloud/shadow mask **kwargs are passed from MaskedCollection.composite() through to
    Sentinel2ClImage._aux_image().
    """
    gd_collection = MaskedCollection.from_list(s2_sr_hm_images)
    comp_ims = []
    for score in [0.3, 0.5]:
        comp_im = gd_collection.composite(mask_method='cloud-score', score=score)
        comp_im._set_region_stats(region_100ha, scale=gd_collection.stats_scale)
        comp_ims.append(comp_im)
    assert comp_ims[0].properties['FILL_PORTION'] > comp_ims[1].properties['FILL_PORTION']


def test_composite_landsat_cloud_mask_params(l8_9_images, region_10000ha):
    """
    Test cloud/shadow mask **kwargs are passed from MaskedCollection.composite() through to
    LandsatImage._aux_image().
    """
    gd_collection = MaskedCollection.from_list(l8_9_images)
    comp_im_wshadows = gd_collection.composite(mask_shadows=False)
    comp_im_wshadows._set_region_stats(region_10000ha, scale=gd_collection.stats_scale)
    comp_im_woshadows = gd_collection.composite(mask_shadows=True)
    comp_im_woshadows._set_region_stats(region_10000ha, scale=gd_collection.stats_scale)
    with_shadows_portion = comp_im_wshadows.properties['FILL_PORTION']
    without_shadows_portion = comp_im_woshadows.properties['FILL_PORTION']
    assert with_shadows_portion > without_shadows_portion


@pytest.mark.parametrize(
    'image_list, method, mask, region, date, cloud_kwargs',
    [
        ('s2_sr_hm_images', CompositeMethod.q_mosaic, True, 'region_100ha', None, {}),
        ('s2_sr_hm_images', CompositeMethod.mosaic, True, None, '2021-10-01', {}),
        ('s2_sr_hm_images', CompositeMethod.medoid, False, None, None, {}),
        ('s2_sr_hm_images', CompositeMethod.median, True, None, None, dict(mask_method='cloud-score', score=0.4)),
        ('l8_9_images', CompositeMethod.q_mosaic, True, 'region_100ha', None, {}),
        ('l8_9_images', CompositeMethod.mosaic, False, None, '2022-03-01', {}),
        ('l8_9_images', CompositeMethod.medoid, True, None, None, dict(mask_cirrus=False, mask_shadows=False)),
        ('l8_9_images', CompositeMethod.median, True, None, None, {}),
        ('l4_5_images', CompositeMethod.q_mosaic, False, 'region_100ha', None, {}),
        (
            'l4_5_images',
            CompositeMethod.mosaic,
            True,
            None,
            '1988-01-01',
            dict(mask_cirrus=False, mask_shadows=False),
        ),
        ('l4_5_images', CompositeMethod.medoid, True, None, None, {}),
        ('l4_5_images', CompositeMethod.median, True, None, None, {}),
        ('gedi_image_list', CompositeMethod.mosaic, True, 'region_100ha', None, {}),
        ('gedi_image_list', CompositeMethod.mosaic, True, None, '2020-09-01', {}),
        ('gedi_image_list', CompositeMethod.medoid, True, None, None, {}),
    ],
)
def test_composite(image_list, method, mask, region, date, cloud_kwargs, request):
    """Test MaskedCollection.composite() runs successfully with a variety of `method` and other parameters."""
    image_list: List = request.getfixturevalue(image_list)
    region: Dict = request.getfixturevalue(region) if region else None
    gd_collection = MaskedCollection.from_list(image_list)
    comp_im = gd_collection.composite(method=method, mask=mask, region=region, date=date, **cloud_kwargs)
    assert comp_im._ee_info is not None and len(comp_im._ee_info) > 0
    assert 'INPUT_IMAGES' in comp_im.properties


def test_composite_errors(gedi_image_list, region_100ha):
    """Test MaskedCollection.composite() error conditions."""
    gedi_collection = MaskedCollection.from_list(gedi_image_list)
    with pytest.raises(ValueError):
        # q-mosaic is only supported on cloud/shadow maskable images
        gedi_collection.composite(method=CompositeMethod.q_mosaic)
    with pytest.raises(ValueError):
        # unknown method
        gedi_collection.composite(method='unknown')
    with pytest.raises(ValueError):
        # date format error
        gedi_collection.composite(method=CompositeMethod.mosaic, date='not a date')
    with pytest.raises(ValueError):
        # composite of empty collection
        empty_collection = gedi_collection.search('2000-01-01', '2000-01-02', region_100ha, 100)
        empty_collection.composite(method=CompositeMethod.mosaic)


@pytest.mark.parametrize('image_list', ['s2_sr_hm_images', 'l8_9_images'])
def test_composite_date(image_list: str, request: pytest.FixtureRequest):
    """Test the composite date is the same as the first input image date."""
    image_list: List = request.getfixturevalue(image_list)
    gd_collection = MaskedCollection.from_list(image_list)
    first_date = datetime.fromtimestamp(
        gd_collection.ee_collection.sort('system:time_start').first().get('system:time_start').getInfo() / 1000,
        tz=timezone.utc,
    )
    comp_im = gd_collection.composite()
    assert comp_im.date == first_date


def test_composite_mult_kwargs(region_100ha):
    """
    When a search filtered collection is composited, test that masks change with different cloud/shadow kwargs i.e.
    test that image *_MASK bands are overwritten in the encapsulated collection.
    """
    gd_collection = MaskedCollection.from_name('COPERNICUS/S2_SR_HARMONIZED')
    filt_collection = gd_collection.search('2022-01-01', '2022-01-10', region_100ha)

    comp_ims = []
    for score in [0.3, 0.5]:
        comp_im = filt_collection.composite(mask_method='cloud-score', score=score)
        comp_im._set_region_stats(region_100ha, scale=filt_collection.stats_scale)
        comp_ims.append(comp_im)

    assert comp_ims[0].properties['FILL_PORTION'] != pytest.approx(comp_ims[1].properties['FILL_PORTION'], abs=1e-1)


@pytest.mark.parametrize('name', ['FAO/WAPOR/2/L1_RET_E', 'MODIS/006/MCD43A4'])
def test_unbounded_search_no_region(name):
    """
    Test searching an unbounded collection without a region raises an exception.
    """
    start_date = '2022-01-01'
    end_date = '2022-01-02'
    gd_collection = MaskedCollection.from_name(name)
    gd_collection = gd_collection.search(start_date, end_date, fill_portion=0)
    with pytest.raises(ValueError) as ex:
        _ = gd_collection.properties
    assert 'unbounded' in str(ex) and 'region' in str(ex)


def test_unknown_collection_error(region_25ha):
    """
    Test searching a non-existent collection gives an error.
    """
    start_date = '2022-01-01'
    end_date = '2022-01-02'
    gd_collection = MaskedCollection.from_name('Unknown')
    with pytest.raises(ee.EEException) as ex:
        gd_collection = gd_collection.search(start_date, end_date, region_25ha)
        _ = gd_collection.properties
    assert 'not found' in str(ex)



"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

import logging
import warnings

import ee

import geedim.schema
from geedim.download import BaseImage
from geedim.enums import CloudMaskMethod, CloudScoreBand
from geedim.utils import get_projection, split_id

logger = logging.getLogger(__name__)

##


class MaskedImage(BaseImage):
    _default_mask = False

    def __init__(self, ee_image: ee.Image, mask: bool = _default_mask, region: dict | ee.Geometry = None, **kwargs):
        """
        A class for describing, masking and downloading an Earth Engine image.

        Parameters
        ----------
        ee_image: ee.Image
            Earth Engine image to encapsulate.
        mask: bool, optional
            Whether to mask the image.
        region: dict, ee.Geometry, optional
            Region in which to find statistics for the image, as a GeoJSON dictionary or ``ee.Geometry``.  Statistics
            are stored in the image properties. If None, statistics are not found (the default).

        **kwargs
            Cloud/shadow masking parameters - see below:
        mask_cirrus: bool, optional
            Whether to mask cirrus clouds.  Valid for Landsat 8-9 images, and for Sentinel-2 images with
            the `qa` ``mask_method``.
        mask_shadows: bool, optional
            Whether to mask cloud shadows.  Valid for Landsat images, and for Sentinel-2 images with
            the `qa` or `cloud-prob` ``mask_method``.
        mask_method: CloudMaskMethod, str, optional
            Method used to mask clouds.  Valid for Sentinel-2 images.  See :class:`~geedim.enums.CloudMaskMethod` for
            details.
        prob: float, optional
            Cloud probability threshold (%). Valid for Sentinel-2 images with the `cloud-prob` ``mask_method``.
        dark: float, optional
            NIR threshold [0-1]. NIR values below this threshold are potential cloud shadows.  Valid for Sentinel-2
            images with the `qa` or `cloud-prob` ``mask_method``.
        shadow_dist: int, optional
            Maximum distance (m) to look for cloud shadows from cloud edges.  Valid for Sentinel-2 images with the `qa`
            or `cloud-prob` ``mask_method``.
        buffer: int, optional
            Distance (m) to dilate cloud/shadow.  Valid for Sentinel-2 images with the `qa` or `cloud-prob`
            ``mask_method``.
        cdi_thresh: float, optional
            Cloud Displacement Index threshold. Values below this threshold are considered potential clouds.
            If this parameter is not specified (=None), the index is not used.  Valid for Sentinel-2 images with the
            `qa` or `cloud-prob` ``mask_method``.  See
            https://developers.google.com/earth-engine/apidocs/ee-algorithms-sentinel2-cdi for details.
        max_cloud_dist: int, optional
            Maximum distance (m) to look for clouds when forming the 'cloud distance' band.  Valid for
            Sentinel-2 images.
        score: float, optional
            Cloud Score+ threshold.  Valid for Sentinel-2 images with the `cloud-score` ``mask_method``.
        cs_band: CloudScoreBand, str, optional
            Cloud Score+ band to threshold.  Valid for Sentinel-2 images with the `cloud-score` ``mask_method``.
        """
        BaseImage.__init__(self, ee_image)
        self._ee_projection = None
        self._add_aux_bands(**kwargs)  # add any mask and cloud distance bands
        if region:
            self._set_region_stats(region)
        if mask:
            self.mask_clouds()

    @property
    def _ee_proj(self) -> ee.Projection:
        """Projection to use for mask calculations and statistics."""
        if self._ee_projection is None:
            # use the minimum scale projection for the base class / generic case (excludes non-fixed projections with
            # 1 deg. scales)
            self._ee_projection = get_projection(self._ee_image, min_scale=True)
        return self._ee_projection

    @staticmethod
    def from_id(image_id: str, **kwargs) -> 'MaskedImage':
        """
        Create a MaskedImage, or sub-class, instance from an Earth Engine image ID.

        Parameters
        ----------
        image_id: str
            ID of the Earth Engine image to encapsulate.
        **kwargs
            Optional keyword arguments to pass to :meth:`__init__`.

        Returns
        -------
        MaskedImage
            A MaskedImage, or sub-class instance.
        """
        cls = class_from_id(image_id)
        ee_image = ee.Image(image_id)
        gd_image = cls(ee_image, **kwargs)
        gd_image._id = image_id  # set the id attribute (avoids a call to getInfo() for .id property)
        return gd_image

    def _aux_image(self, **kwargs) -> ee.Image:
        """
        Retrieve the auxiliary image (MaskedImage provides an image with a FILL_MASK band only).

        Derived classes should override this method and return an image with FILL_MASK and any other auxiliary bands
        they support.
        """
        return self.ee_image.mask().reduce(ee.Reducer.allNonZero()).rename('FILL_MASK')

    def _add_aux_bands(self, **kwargs):
        """
        Add auxiliary bands to the encapsulated image. Existing auxiliary bands are overwritten, unless the image has
        no fixed projection and existing auxiliary bands (i.e. is likely a MaskedCollection composite image).
        """
        aux_image = self._aux_image(**kwargs)
        proj = self.ee_image.select(0).projection()
        has_fixed_scale = proj.nominalScale().toInt64().neq(111319)  # 1 deg in meters
        has_no_aux_bands = ee.Number(self.ee_image.bandNames().indexOf('FILL_MASK').lt(0))
        # overwrite unless it is a composite image with existing aux bands
        overwrite = has_no_aux_bands.Or(has_fixed_scale)
        self.ee_image = ee.Image(
            ee.Algorithms.If(overwrite, self.ee_image.addBands(aux_image, overwrite=True), self.ee_image)
        )

    def _set_region_stats(self, region: dict | ee.Geometry = None, scale: float = None):
        """
        Set FILL_PORTION and CLOUDLESS_PORTION on the encapsulated image for the specified region.  Derived classes
        should override this method and set CLOUDLESS_PORTION, and/or other statistics they support.

        Parameters
        ----------
        region: dict, ee.Geometry, optional
            Region in which to find statistics for the image, as a GeoJSON dictionary or ``ee.Geometry``.  If not
            specified, the image footprint is used.
        scale: float, optional
            Re-project to this scale when finding statistics.  Defaults to the scale of
            :attr:`~MaskedImage._ee_proj`.  Should be provided if the encapsulated image is a composite / without a
            fixed projection.
        """
        if not region:
            region = self.ee_image.geometry()  # use the image footprint

        # composite images have no fixed projection (scale = 1deg) and need the scale kwarg
        scale = scale or self._ee_proj.nominalScale()

        # Find the fill portion as the (sum over the region of FILL_MASK) divided by (sum over the region of a constant
        # image (==1)).  Note that a mean reducer, does not find the mean over the region, but the mean over the part
        # of the region covered by the image.
        stats_image = ee.Image([self.ee_image.select('FILL_MASK').unmask(), ee.Image(1).rename('REGION_SUM')])
        # Note: sometimes proj has no EPSG in crs(), hence use crs=proj and not crs=proj.crs() below
        sums = stats_image.reduceRegion(
            reducer="sum", geometry=region, crs=self._ee_proj, scale=scale, bestEffort=True, maxPixels=1e6
        )

        fill_portion = ee.Number(sums.get('FILL_MASK')).divide(ee.Number(sums.get('REGION_SUM'))).multiply(100)

        # set the encapsulated image properties
        self.ee_image = self.ee_image.set('FILL_PORTION', fill_portion)
        # set CLOUDLESS_PORTION=100 for the generic case, where cloud/shadow masking is not supported
        self.ee_image = self.ee_image.set('CLOUDLESS_PORTION', 100.0)

    def mask_clouds(self):
        """Apply the cloud/shadow mask if supported, otherwise apply the fill mask."""
        self.ee_image = self.ee_image.updateMask(self.ee_image.select('FILL_MASK'))


class CloudMaskedImage(MaskedImage):
    """A base class for encapsulating cloud/shadow masked images."""

    def _cloud_dist(self, cloudless_mask: ee.Image = None, max_cloud_dist: float = 5000) -> ee.Image:
        """Find the cloud/shadow distance in units of 10m."""
        if not cloudless_mask:
            cloudless_mask = self.ee_image.select('CLOUDLESS_MASK')

        cloud_shadow_mask = cloudless_mask.Not()
        cloud_pix = ee.Number(max_cloud_dist).divide(self._ee_proj.nominalScale()).round()  # cloud_dist in pixels

        # Find distance to nearest cloud/shadow (units of 10m).  Distances are found for all pixels, including masked /
        # invalid pixels, which are treated as 0 (non cloud/shadow).
        cloud_dist = (
            cloud_shadow_mask.fastDistanceTransform(neighborhood=cloud_pix, units='pixels', metric='squared_euclidean')
            .sqrt()
            .multiply(self._ee_proj.nominalScale().divide(10))
        )

        # Reproject to force calculation at correct scale.
        cloud_dist = cloud_dist.reproject(crs=self._ee_proj, scale=self._ee_proj.nominalScale()).rename('CLOUD_DIST')

        # Prevent use of invalid pixels.
        cloud_dist = cloud_dist.updateMask(cloudless_mask.mask())

        # Clip cloud_dist to max_cloud_dist.
        cloud_dist = cloud_dist.where(cloud_dist.gt(ee.Image(max_cloud_dist / 10)), max_cloud_dist / 10)

        # cloud_dist is float64 by default, so convert to Uint16 here to avoid forcing the whole image to float64 on
        # download.
        return cloud_dist.toUint16().rename('CLOUD_DIST')

    def _set_region_stats(self, region: dict | ee.Geometry = None, scale: float = None):
        """
        Set FILL_PORTION and CLOUDLESS_PORTION on the encapsulated image for the specified region.

        Parameters
        ----------
        region: dict, ee.Geometry, optional
            Region in which to find statistics for the image, as a GeoJSON dictionary or ``ee.Geometry``.  If not
            specified, the image footprint is used.
        scale: float, optional
            Re-project to this scale when finding statistics.  Defaults to the scale of
            :attr:`~MaskedImage._ee_proj`.  Should be provided if the encapsulated image is a composite / without a
            fixed projection.
        """
        if not region:
            region = self.ee_image.geometry()  # use the image footprint

        # composite images have no fixed projection (scale = 1deg) and need the scale kwarg
        scale = scale or self._ee_proj.nominalScale()

        # Find the fill portion as the (sum over the region of FILL_MASK) divided by (sum over the region of a
        # constant image (==1)).  Then find cloudless portion as portion of fill that is cloudless.  Note that a mean
        # reducer, does not find the mean over the region, but the mean over the part of the region covered by the
        # image.
        stats_image = ee.Image(
            [self.ee_image.select(['FILL_MASK', 'CLOUDLESS_MASK']).unmask(), ee.Image(1).rename('REGION_SUM')]
        )
        sums = stats_image.reduceRegion(
            reducer="sum", geometry=region, crs=self._ee_proj, scale=scale, bestEffort=True, maxPixels=1e6
        ).rename(['FILL_MASK', 'CLOUDLESS_MASK'], ['FILL_PORTION', 'CLOUDLESS_PORTION'])
        fill_portion = ee.Number(sums.get('FILL_PORTION')).divide(ee.Number(sums.get('REGION_SUM'))).multiply(100)
        cloudless_portion = (
            ee.Number(sums.get('CLOUDLESS_PORTION')).divide(ee.Number(sums.get('FILL_PORTION'))).multiply(100)
        )

        # set the encapsulated image properties
        region_stats = ee.Dictionary(dict(FILL_PORTION=fill_portion, CLOUDLESS_PORTION=cloudless_portion))
        self.ee_image = self.ee_image.set(region_stats)

    def _aux_image(self, **kwargs) -> ee.Image:
        """
        Retrieve the auxiliary image containing cloud/shadow masks and cloud distance.  The returned image should
        contain at least FILL_MASK, CLOUDLESS_MASK and CLOUD_DIST bands.

        See :meth:`LandsatImage._aux_image` for an example.
        """
        raise NotImplementedError('This virtual method should be overridden by derived classes.')

    def mask_clouds(self):
        """Apply the cloud/shadow mask."""
        self.ee_image = self.ee_image.updateMask(self.ee_image.select('CLOUDLESS_MASK'))


class LandsatImage(CloudMaskedImage):
    """
    Class for cloud/shadow masking of Landsat level 2, collection 2 images.

    Supports images from:
    * LANDSAT/LT04/C02/T1_L2
    * LANDSAT/LT05/C02/T1_L2
    * LANDSAT/LE07/C02/T1_L2
    * LANDSAT/LC08/C02/T1_L2
    * LANDSAT/LC09/C02/T1_L2
    """

    @property
    def _ee_proj(self) -> ee.Projection:
        if self._ee_projection is None:
            # use the default projection (all landsat bands have same scale)
            self._ee_projection = self._ee_image.projection()
        return self._ee_projection

    def _aux_image(self, mask_shadows: bool = True, mask_cirrus: bool = True, max_cloud_dist: int = 5000) -> ee.Image:
        """
        Retrieve the auxiliary image containing cloud/shadow masks and cloud distance.

        Parameters
        ----------
        mask_shadows: bool, optional
            Whether to mask cloud shadows.
        mask_cirrus: bool, optional
            Whether to mask cirrus clouds.  Valid for Landsat 8-9 images.
        max_cloud_dist: int, optional
            Maximum distance (m) to look for clouds when forming the 'cloud distance' band.  Valid for
            Sentinel-2 images.

        Returns
        -------
        ee.Image
            An Earth Engine image containing *_MASK and CLOUD_DIST bands.
        """
        ee_image = self._ee_image
        qa_pixel = ee_image.select('QA_PIXEL')

        # construct fill mask from Earth Engine mask and QA_PIXEL
        ee_mask = ee_image.select('SR_B.*').mask().reduce(ee.Reducer.allNonZero())
        fill_mask = qa_pixel.bitwiseAnd(1).eq(0).And(ee_mask).rename('FILL_MASK')

        shadow_mask = qa_pixel.bitwiseAnd(0b10000).neq(0).rename('SHADOW_MASK')
        if mask_cirrus:
            cloud_mask = qa_pixel.bitwiseAnd(0b1100).neq(0).rename('CLOUD_MASK')
        else:
            cloud_mask = qa_pixel.bitwiseAnd(0b1000).neq(0).rename('CLOUD_MASK')

        # combine cloud, shadow and fill masks into cloudless mask
        cloud_shadow_mask = (cloud_mask.Or(shadow_mask)) if mask_shadows else cloud_mask
        cloudless_mask = cloud_shadow_mask.Not().And(fill_mask).rename('CLOUDLESS_MASK')

        # copy cloud distance from existing ST_CDIST band (in 10m units), and clip to max_cloud_dist
        cloud_dist = ee_image.select('ST_CDIST').rename('CLOUD_DIST').toUint16()
        cloud_dist = cloud_dist.where(cloud_dist.gt(ee.Image(max_cloud_dist / 10)), max_cloud_dist / 10)

        return ee.Image([fill_mask, cloud_mask, shadow_mask, cloudless_mask, cloud_dist])


class Sentinel2ClImage(CloudMaskedImage):
    """Base class for cloud/shadow masking of Sentinel-2 TOA and SR images."""

    @property
    def _ee_proj(self) -> ee.Projection:
        if self._ee_projection is None:
            # use the B1 projection with maximum scale (60m) to reduce processing times (some S2 images have empty QA
            # bands with no fixed projection, so utils.get_projection(min_scale=False) should not be used here).
            self._ee_projection = self._ee_image.select(0).projection()
        return self._ee_projection

    def _aux_image(
        self,
        s2_toa: bool = False,
        mask_cirrus: bool = True,
        mask_shadows: bool = True,
        mask_method: CloudMaskMethod = CloudMaskMethod.cloud_score,
        prob: float = 60,
        dark: float = 0.15,
        shadow_dist: int = 1000,
        buffer: int = 50,
        cdi_thresh: float = None,
        max_cloud_dist: int = 5000,
        score: float = 0.6,
        cs_band: CloudScoreBand = CloudScoreBand.cs,
    ) -> ee.Image:
        """Derive cloud, shadow and validity masks for the encapsulated image.

        Parts adapted from https://github.com/r-earthengine/ee_extra, under Apache 2.0 license.
        """
        mask_method = CloudMaskMethod(mask_method)
        if mask_method is not CloudMaskMethod.cloud_score:
            warnings.warn(
                f"The '{mask_method}' mask method is deprecated and will be removed in a future release.  Please "
                f"switch to 'cloud-score'.",
                category=FutureWarning,
            )

        def match_image(ee_image: ee.Image, collection: str, band: str, match_prop: str = 'system:index') -> ee.Image:
            """Return an image from ``collection`` matching ``ee_image`` with single ``band`` selected, or a fully
            masked image if no match is found.
            """
            # default fully masked image
            default = ee.Image().updateMask(0)
            default = default.rename(band)

            # find matching image
            filt = ee.Filter.eq(match_prop, ee_image.get(match_prop))
            match = ee.ImageCollection(collection).filter(filt).first()

            # revert to default if no match found
            match = ee.Image(ee.List([match, default]).reduce(ee.Reducer.firstNonNull()))
            return match.select(band)

        def cloud_cast_shadow_mask(ee_image: ee.Image, cloud_mask: ee.Image) -> ee.Image:
            """Create & return a shadow mask for ``ee_image`` by projecting shadows from ``cloud_mask``.

            Adapted from https://developers.google.com/earth-engine/tutorials/community/sentinel-2-s2cloudless.
            """
            dark_mask = ee_image.select('B8').lt(dark * 1e4)
            if not s2_toa:
                # exclude water
                dark_mask = ee_image.select('SCL').neq(6).And(dark_mask)

            # Find shadow direction.  Note that the angle convention depends on the reproject args below.
            shadow_azimuth = ee.Number(90).subtract(ee.Number(ee_image.get('MEAN_SOLAR_AZIMUTH_ANGLE')))

            # Project the cloud mask in the shadow direction (can project the mask into invalid areas).
            proj_pixels = ee.Number(shadow_dist).divide(self._ee_proj.nominalScale()).round()
            cloud_cast_proj = cloud_mask.directionalDistanceTransform(shadow_azimuth, proj_pixels).select('distance')

            # Reproject to force calculation at the correct scale.
            cloud_cast_mask = cloud_cast_proj.mask().reproject(crs=self._ee_proj, scale=self._ee_proj.nominalScale())

            # Remove any projections in invalid areas.
            cloud_cast_mask = cloud_cast_mask.updateMask(cloud_mask.mask())

            # Find shadow mask as intersection between projected clouds and dark areas.
            shadow_mask = cloud_cast_mask.And(dark_mask)

            return shadow_mask.rename('SHADOW_MASK')

        def qa_cloud_mask(ee_image: ee.Image) -> ee.Image:
            """Create & return a cloud mask for ``ee_image`` using the QA60 band."""
            # mask QA60 if it is between Feb 2022 - Feb 2024 to invalidate the cloud mask (QA* bands are not populated
            # over this period and may be masked / transparent or zero).
            qa_valid = (
                ee_image.date()
                .difference(ee.Date('2022-02-01'), 'days')
                .lt(0)
                .Or(ee_image.date().difference(ee.Date('2024-02-01'), 'days').gt(0))
            )
            qa = ee_image.select('QA60').updateMask(qa_valid)

            cloud_mask = qa.bitwiseAnd(1 << 10).neq(0)
            if mask_cirrus:
                cloud_mask = cloud_mask.Or(qa.bitwiseAnd(1 << 11).neq(0))

            return cloud_mask.rename('CLOUD_MASK')

        def cloud_prob_cloud_mask(ee_image: ee.Image) -> tuple[ee.Image, ee.Image]:
            """Return the cloud probability thresholded cloud mask, and cloud probability image for ``ee_image``."""
            cloud_prob = match_image(ee_image, 'COPERNICUS/S2_CLOUD_PROBABILITY', 'probability')
            cloud_mask = cloud_prob.gte(prob)
            return cloud_mask.rename('CLOUD_MASK'), cloud_prob.rename('CLOUD_PROB')

        def cloud_score_cloud_shadow_mask(ee_image: ee.Image) -> tuple[ee.Image, ee.Image]:
            """Return the cloud score thresholded cloud mask, and cloud score image for ``ee_image``."""
            cloud_score = match_image(ee_image, 'GOOGLE/CLOUD_SCORE_PLUS/V1/S2_HARMONIZED', cs_band.name)
            cloud_shadow_mask = cloud_score.lte(score)
            return cloud_shadow_mask.rename('CLOUD_SHADOW_MASK'), cloud_score.rename('CLOUD_SCORE')

        def cdi_cloud_mask(ee_image: ee.Image) -> ee.Image:
            """Return the CDI cloud mask for ``ee_image``.
            See https://developers.google.com/earth-engine/apidocs/ee-algorithms-sentinel2-cdi for detail.
            """
            if s2_toa:
                s2_toa_image = ee_image
            else:
                # get the Sentinel-2 TOA image that corresponds to ``ee_image``
                idx = ee_image.get('system:index')
                s2_toa_image = ee.ImageCollection('COPERNICUS/S2').filter(ee.Filter.eq('system:index', idx)).first()
            cdi_image = ee.Algorithms.Sentinel2.CDI(s2_toa_image)
            return cdi_image.lt(cdi_thresh).rename('CDI_CLOUD_MASK')

        def cloud_shadow_masks(ee_image: ee.Image) -> dict[str, ee.Image]:
            """Return a dictionary of cloud/shadow masks & related images for ``ee_image``."""
            aux_bands = {}
            if mask_method is CloudMaskMethod.cloud_score:
                aux_bands['cloud_shadow'], aux_bands['score'] = cloud_score_cloud_shadow_mask(ee_image)
            else:
                if mask_method is CloudMaskMethod.qa:
                    aux_bands['cloud'] = qa_cloud_mask(ee_image)
                else:
                    aux_bands['cloud'], aux_bands['prob'] = cloud_prob_cloud_mask(ee_image)

                if cdi_thresh is not None:
                    aux_bands['cloud'] = aux_bands['cloud'].And(cdi_cloud_mask(ee_image))

                aux_bands['shadow'] = cloud_cast_shadow_mask(ee_image, aux_bands['cloud'])

                if mask_shadows:
                    aux_bands['cloud_shadow'] = aux_bands['cloud'].Or(aux_bands['shadow'])
                else:
                    aux_bands['cloud_shadow'] = aux_bands['cloud']
                    aux_bands.pop('shadow', None)

                # do a morphological opening that removes small (20m) blobs from the mask and then dilates
                aux_bands['cloud_shadow'] = (
                    aux_bands['cloud_shadow'].focal_min(20, units='meters').focal_max(buffer, units='meters')
                )
            return aux_bands

        # get cloud/shadow etc bands
        aux_bands = cloud_shadow_masks(self.ee_image)

        # derive a fill mask from the Earth Engine mask for the surface reflectance bands
        aux_bands['fill'] = self.ee_image.select('B.*').mask().reduce(ee.Reducer.allNonZero())

        # clip fill mask to the image footprint (without this step we get memory limit errors on download)
        aux_bands['fill'] = aux_bands['fill'].clip(self.ee_image.geometry()).rename('FILL_MASK')

        # combine masks into cloudless_mask
        aux_bands['cloudless'] = (aux_bands['cloud_shadow'].Not()).And(aux_bands['fill']).rename('CLOUDLESS_MASK')
        aux_bands.pop('cloud_shadow')

        # construct and return the auxiliary image
        aux_bands['dist'] = self._cloud_dist(cloudless_mask=aux_bands['cloudless'], max_cloud_dist=max_cloud_dist)
        return ee.Image(list(aux_bands.values()))


class Sentinel2SrClImage(Sentinel2ClImage):
    """Class for cloud/shadow masking of Sentinel-2 SR (COPERNICUS/S2_SR & COPERNICUS/S2_SR_HARMONIZED) images."""

    def _aux_image(self, s2_toa: bool = False, **kwargs) -> ee.Image:
        return Sentinel2ClImage._aux_image(self, s2_toa=False, **kwargs)


class Sentinel2ToaClImage(Sentinel2ClImage):
    """Class for cloud/shadow masking of Sentinel-2 TOA (COPERNICUS/S2 & COPERNICUS/S2_HARMONIZED) images."""

    def _aux_image(self, s2_toa: bool = False, **kwargs) -> ee.Image:
        return Sentinel2ClImage._aux_image(self, s2_toa=True, **kwargs)


def class_from_id(image_id: str) -> type:
    """Return the *Image class that corresponds to the provided Earth Engine image/collection ID."""
    ee_coll_name, _ = split_id(image_id)
    if image_id in geedim.schema.collection_schema:
        return geedim.schema.collection_schema[image_id]['image_type']
    elif ee_coll_name in geedim.schema.collection_schema:
        return geedim.schema.collection_schema[ee_coll_name]['image_type']
    else:
        return MaskedImage



"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import io
import json
import logging
from collections import namedtuple

import ee
import numpy as np
import pytest
import rasterio as rio
import requests
from rasterio import Affine
from rasterio.windows import Window
from tqdm.auto import tqdm

from geedim.errors import TileError
from geedim.tile import Tile
from geedim.utils import retry_session

BaseImageLike = namedtuple('BaseImageLike', ['ee_image', 'crs', 'transform', 'shape', 'count', 'dtype'])


@pytest.fixture(scope='module')
def mock_base_image(region_25ha: dict) -> BaseImageLike:
    """A BaseImage mock containing a synthetic ee.Image."""
    ee_image = ee.Image([1, 2, 3]).reproject(crs='EPSG:4326', scale=30).clip(region_25ha)
    ee_info = ee_image.getInfo()
    band_info = ee_info['bands'][0]
    transform = Affine(*band_info['crs_transform']) * Affine.translation(*band_info['origin'])
    return BaseImageLike(ee_image, 'EPSG:3857', transform, tuple(band_info['dimensions'][::-1]), 3, 'uint8')


@pytest.fixture(scope='module')
def synth_tile(mock_base_image: BaseImageLike) -> Tile:
    """A tile representing the whole of ``mock_base_image``."""
    window = Window(0, 0, *mock_base_image.shape[::-1])
    return Tile(mock_base_image, window)


@pytest.fixture(scope='function')
def mock_ee_image(monkeypatch: pytest.MonkeyPatch) -> None:
    """Patch ee.Image.getDownloadURL() to return None."""

    def getDownloadURL(*args, **kwargs):
        return None

    monkeypatch.setattr(ee.Image, 'getDownloadURL', getDownloadURL)


@pytest.fixture(scope='module')
def gtiff_bytes(mock_base_image: BaseImageLike) -> bytes:
    """GeoTIFF bytes for ``mock_base_image``."""
    array = np.ones((mock_base_image.count, *mock_base_image.shape)) * np.array([1, 2, 3]).reshape(-1, 1, 1)

    buf = io.BytesIO()
    with rio.open(
        buf,
        'w',
        **rio.default_gtiff_profile,
        width=mock_base_image.shape[1],
        height=mock_base_image.shape[0],
        count=mock_base_image.count,
    ) as ds:
        ds.write(array)

    buf.seek(0)
    return buf.read()


def test_create(mock_base_image: BaseImageLike):
    """Test creation of a Tile object."""
    window = Window(0, 0, *mock_base_image.shape[::-1])
    tile = Tile(mock_base_image, window)
    assert tile.window == window
    assert tile._transform == mock_base_image.transform
    assert tile._shape == mock_base_image.shape


@pytest.mark.parametrize('session', [None, retry_session()])
def test_download(synth_tile: Tile, session):
    """Test downloading the synthetic image tile."""
    dtype_size = np.dtype(synth_tile._exp_image.dtype).itemsize
    raw_download_size = synth_tile._shape[0] * synth_tile._shape[1] * synth_tile._exp_image.count * dtype_size
    bar = tqdm(total=float(raw_download_size))
    array = synth_tile.download(session=session, bar=bar)

    assert array is not None
    assert array.shape == (synth_tile._exp_image.count, *synth_tile._exp_image.shape)
    assert array.dtype == np.dtype(synth_tile._exp_image.dtype)
    for i in range(array.shape[0]):
        assert np.all(array[i] == i + 1)
    assert bar.n == pytest.approx(raw_download_size, rel=0.01)


def test_mem_limit_error(synth_tile: Tile, mock_ee_image: None):
    """Test downloading raises the 'user memory limit exceeded' error with a mock response."""
    # patch session.get() to return a mock response with EE memory limit error
    session = retry_session()
    msg = 'User memory limit exceeded.'

    def get(url, **kwargs):
        response = requests.Response()
        response.status_code = 400
        response.headers = {'content-length': '1'}
        response._content = json.dumps({'error': {'message': msg}}).encode()
        return response

    session.get = get

    # test memory limit error is raised on download
    with pytest.raises(TileError) as ex:
        synth_tile.download(session=session)
    assert msg in str(ex.value)


def test_retry(synth_tile: Tile, mock_ee_image: None, gtiff_bytes: bytes, caplog: pytest.LogCaptureFixture):
    """Test downloading retries invalid tiles until it succeeds."""
    # create progress bar
    dtype_size = np.dtype(synth_tile._exp_image.dtype).itemsize
    raw_download_size = synth_tile._shape[0] * synth_tile._shape[1] * synth_tile._exp_image.count * dtype_size
    bar = tqdm(total=float(raw_download_size))

    # create mock invalid responses for each retry
    responses = []
    for _ in range(5):
        response = requests.Response()
        response.status_code = 200
        response.headers = {'content-length': str(len(gtiff_bytes))}
        response.raw = io.BytesIO(b'error')
        responses.append(response)

    # make the last response valid
    responses[-1].raw = io.BytesIO(gtiff_bytes)

    # patch session.get() to pop and return a mocked response from the list
    session = retry_session()

    def get(url, **kwargs):
        return responses.pop(0)

    session.get = get

    # test the tile is downloaded correctly, after retries
    with caplog.at_level(logging.WARNING):
        array = synth_tile.download(session=session, bar=bar, backoff_factor=0)

    assert array.shape == (synth_tile._exp_image.count, *synth_tile._exp_image.shape)
    assert array.dtype == np.dtype(synth_tile._exp_image.dtype)
    for i in range(array.shape[0]):
        assert np.all(array[i] == i + 1)

    # test progress bar is adjusted for retries
    assert bar.n == pytest.approx(raw_download_size, rel=0.01)

    # test retry logs
    assert 'retry' in caplog.text and 'not recognized' in caplog.text


def test_retry_error(synth_tile: Tile, mock_ee_image: None, gtiff_bytes: bytes):
    """Test downloading raises an error when the maximum retries are reached."""
    # create progress bar
    dtype_size = np.dtype(synth_tile._exp_image.dtype).itemsize
    raw_download_size = synth_tile._shape[0] * synth_tile._shape[1] * synth_tile._exp_image.count * dtype_size
    bar = tqdm(total=float(raw_download_size))

    # patch session.get() to return a mock response with invalid bytes
    session = retry_session()

    def get(url, **kwargs):
        response = requests.Response()
        response.status_code = 200
        response.headers = {'content-length': '10'}
        response.raw = io.BytesIO(b'error')
        return response

    session.get = get

    # test max retries error is raised on download
    with pytest.raises(TileError) as ex:
        synth_tile.download(session=session, bar=bar, backoff_factor=0)
    assert 'maximum retries' in str(ex.value)

    # test progress bar is adjusted for retries
    assert bar.n == pytest.approx(0)



"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from typing import Dict, Iterable

import ee
import numpy as np
import pytest
import rasterio as rio

from geedim import CloudMaskMethod
from geedim.enums import CloudScoreBand
from geedim.mask import class_from_id, CloudMaskedImage, MaskedImage


def test_class_from_id(landsat_image_ids, s2_sr_image_id, s2_toa_hm_image_id, generic_image_ids):
    """Test class_from_id()."""
    from geedim.mask import LandsatImage, Sentinel2SrClImage, Sentinel2ToaClImage

    assert all([class_from_id(im_id) == LandsatImage for im_id in landsat_image_ids])
    assert all([class_from_id(im_id) == MaskedImage for im_id in generic_image_ids])
    assert class_from_id(s2_sr_image_id) == Sentinel2SrClImage
    assert class_from_id(s2_toa_hm_image_id) == Sentinel2ToaClImage


def test_from_id():
    """Test MaskedImage.from_id() sets _id."""
    ee_id = 'MODIS/006/MCD43A4/2022_01_01'
    gd_image = MaskedImage.from_id(ee_id)
    assert gd_image._id == ee_id


@pytest.mark.parametrize(
    'masked_image',
    [
        'user_masked_image',
        'modis_nbar_masked_image',
        'gch_masked_image',
        's1_sar_masked_image',
        'gedi_agb_masked_image',
        'gedi_cth_masked_image',
        'landsat_ndvi_masked_image',
    ],
)
def test_gen_aux_bands_exist(masked_image: str, request: pytest.FixtureRequest):
    """Test the presence of auxiliary band (i.e. FILL_MASK) in generic masked images."""
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    band_names = masked_image.ee_image.bandNames().getInfo()
    assert 'FILL_MASK' in band_names


@pytest.mark.parametrize(
    'masked_image',
    [
        's2_sr_masked_image',
        's2_toa_masked_image',
        's2_sr_hm_masked_image',
        's2_toa_hm_masked_image',
        'l9_masked_image',
        'l8_masked_image',
        'l7_masked_image',
        'l5_masked_image',
        'l4_masked_image',
    ],
)
def test_cloud_mask_aux_bands_exist(masked_image: str, request: pytest.FixtureRequest):
    """Test the presence of auxiliary bands in cloud masked images."""
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    band_names = masked_image.ee_image.bandNames().getInfo()
    exp_band_names = {'FILL_MASK', 'CLOUDLESS_MASK', 'CLOUD_DIST'}
    assert exp_band_names.intersection(band_names) == exp_band_names


@pytest.mark.parametrize(
    'masked_image',
    [
        's2_sr_masked_image',
        's2_toa_masked_image',
        's2_sr_hm_masked_image',
        's2_toa_hm_masked_image',
        'l9_masked_image',
        'l8_masked_image',
        'l7_masked_image',
        'l5_masked_image',
        'l4_masked_image',
        'user_masked_image',
        'modis_nbar_masked_image',
        'gch_masked_image',
        's1_sar_masked_image',
        'gedi_agb_masked_image',
        'gedi_cth_masked_image',
        'landsat_ndvi_masked_image',
    ],
)
def test_set_region_stats(masked_image: str, region_100ha, request: pytest.FixtureRequest):
    """
    Test MaskedImage._set_region_stats() generates the expected properties and that these are in the valid range.
    """
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    masked_image._set_region_stats(region_100ha, scale=masked_image._ee_proj.nominalScale())
    for stat_name in ['FILL_PORTION', 'CLOUDLESS_PORTION']:
        assert stat_name in masked_image.properties
        assert masked_image.properties[stat_name] >= 0 and masked_image.properties[stat_name] <= 100


@pytest.mark.parametrize(
    'masked_image, exp_scale',
    [
        ('s2_sr_hm_masked_image', 60),
        ('l9_masked_image', 30),
        ('l4_masked_image', 30),
        ('s1_sar_masked_image', 10),
        ('gedi_agb_masked_image', 1000),
        # include fixtures with bands that have no fixed projection
        ('s2_sr_hm_qa_zero_masked_image', 60),
        ('s2_sr_hm_nocp_masked_image', 60),
        ('s2_sr_hm_nocs_masked_image', 60),
    ],
)
def test_ee_proj(masked_image: str, exp_scale: float, request: pytest.FixtureRequest):
    """Test MaskedImage._ee_proj has the correct scale and CRS."""
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    proj = masked_image._ee_proj.getInfo()

    assert np.abs(proj['transform'][0]) == pytest.approx(exp_scale, rel=1e-3)
    assert proj.get('crs', 'wkt') != 'EPSG:4326'


@pytest.mark.parametrize('image_id', ['l9_image_id', 'l8_image_id', 'l7_image_id', 'l5_image_id', 'l4_image_id'])
def test_landsat_cloudless_portion(image_id: str, request: pytest.FixtureRequest):
    """Test `geedim` CLOUDLESS_PORTION for the whole image against related Landsat CLOUD_COVER property."""
    image_id: MaskedImage = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, mask_shadows=False, mask_cirrus=False)
    masked_image._set_region_stats()

    # landsat provided cloudless portion
    landsat_cloudless_portion = 100 - float(masked_image.properties['CLOUD_COVER'])
    assert masked_image.properties['CLOUDLESS_PORTION'] == pytest.approx(landsat_cloudless_portion, abs=5)


@pytest.mark.parametrize('image_id', ['s2_toa_image_id', 's2_sr_image_id', 's2_toa_hm_image_id', 's2_sr_hm_image_id'])
def test_s2_cloudless_portion(image_id: str, request: pytest.FixtureRequest):
    """Test `geedim` CLOUDLESS_PORTION for the whole image against CLOUDY_PIXEL_PERCENTAGE Sentinel-2 property."""
    # Note that CLOUDY_PIXEL_PERCENTAGE does not use Cloud Score+ data and does not include shadows, which Cloud Score+
    # does.  So CLOUDLESS_PORTION (with cloud-score method) will only roughly match CLOUDY_PIXEL_PERCENTAGE.
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, mask_method='cloud-score')
    masked_image._set_region_stats()

    # S2 provided cloudless portion
    s2_cloudless_portion = 100 - float(masked_image.properties['CLOUDY_PIXEL_PERCENTAGE'])
    assert masked_image.properties['CLOUDLESS_PORTION'] == pytest.approx(s2_cloudless_portion, abs=10)


@pytest.mark.parametrize('image_id', ['l9_image_id'])
def test_landsat_cloudmask_params(image_id: str, request: pytest.FixtureRequest):
    """Test Landsat cloud/shadow masking `mask_shadows` and `mask_cirrus` parameters."""
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, mask_shadows=False, mask_cirrus=False)
    masked_image._set_region_stats()
    # cloud-free portion
    cloud_only_portion = 100 * masked_image.properties['CLOUDLESS_PORTION']
    masked_image = MaskedImage.from_id(image_id, mask_shadows=True, mask_cirrus=False)
    masked_image._set_region_stats()
    # cloud and shadow-free portion
    cloud_shadow_portion = 100 * masked_image.properties['CLOUDLESS_PORTION']
    masked_image = MaskedImage.from_id(image_id, mask_shadows=True, mask_cirrus=True)
    masked_image._set_region_stats()
    # cloud, cirrus and shadow-free portion
    cloudless_portion = 100 * masked_image.properties['CLOUDLESS_PORTION']

    # test `mask_shadows` and `mask_cirrus` affect CLOUDLESS_PORTION as expected
    assert cloud_only_portion > cloud_shadow_portion
    assert cloud_shadow_portion > cloudless_portion


@pytest.mark.parametrize('image_id', ['s2_sr_hm_image_id', 's2_toa_hm_image_id'])
def test_s2_cloudmask_mask_shadows(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """Test S2 cloud/shadow masking `mask_shadows` parameter."""
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, mask_method='cloud-prob', mask_shadows=False)
    masked_image._set_region_stats(region_10000ha)
    # cloud-free portion
    cloud_only_portion = 100 * masked_image.properties['CLOUDLESS_PORTION']
    masked_image = MaskedImage.from_id(image_id, mask_method='cloud-prob', mask_shadows=True)
    masked_image._set_region_stats(region_10000ha)
    # cloud and shadow-free portion
    cloudless_portion = 100 * masked_image.properties['CLOUDLESS_PORTION']
    assert cloud_only_portion > cloudless_portion


@pytest.mark.parametrize('image_id', ['s2_sr_hm_image_id', 's2_toa_hm_image_id'])
def test_s2_cloudmask_prob(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """Test S2 cloud/shadow masking `prob` parameter with the `cloud-prob` method."""
    image_id: str = request.getfixturevalue(image_id)
    cl_portions = []
    for prob in [80, 40]:
        masked_image = MaskedImage.from_id(image_id, mask_shadows=True, prob=prob, mask_method='cloud-prob')
        masked_image._set_region_stats(region_10000ha)
        cl_portions.append(100 * masked_image.properties['CLOUDLESS_PORTION'])
    # test there is more cloud (less CLOUDLESS_PORTION) with prob=40 as compared to prob=80
    assert cl_portions[0] > cl_portions[1] > 0


@pytest.mark.parametrize('image_id', ['s2_sr_hm_image_id', 's2_toa_hm_image_id'])
def test_s2_cloudmask_score(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """Test S2 cloud/shadow masking `score` parameter with the `cloud-score` method."""
    image_id: str = request.getfixturevalue(image_id)
    cl_portions = []
    for score in [0.6, 0.3]:
        masked_image = MaskedImage.from_id(image_id, mask_shadows=True, score=score, mask_method='cloud-score')
        masked_image._set_region_stats(region_10000ha)
        cl_portions.append(100 * masked_image.properties['CLOUDLESS_PORTION'])
    # test there is more cloud (less CLOUDLESS_PORTION) with score=0.3 as compared to score=0.6
    assert cl_portions[0] < cl_portions[1] > 0


@pytest.mark.parametrize('image_id', ['s2_sr_hm_image_id', 's2_toa_hm_image_id'])
def test_s2_cloudmask_cs_band(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """Test S2 cloud/shadow masking `cs_band` parameter with the `cloud-score` method."""
    image_id: str = request.getfixturevalue(image_id)
    cl_portions = []
    for cs_band in CloudScoreBand:
        masked_image = MaskedImage.from_id(image_id, mask_method='cloud-score', cs_band=cs_band)
        masked_image._set_region_stats(region_10000ha)
        cl_portions.append(100 * masked_image.properties['CLOUDLESS_PORTION'])

    # test `cs_band` changes CLOUDLESS_PORTION but not by much
    assert len(set(cl_portions)) == len(cl_portions)
    assert all([cl_portions[0] != pytest.approx(cp, abs=10) for cp in cl_portions[1:]])
    assert all([cp != pytest.approx(0, abs=1) for cp in cl_portions])


@pytest.mark.parametrize('image_id', ['s2_sr_hm_image_id', 's2_toa_hm_image_id'])
def test_s2_cloudmask_method(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """Test S2 cloud/shadow masking `mask_method` parameter."""
    image_id: str = request.getfixturevalue(image_id)
    cl_portions = []
    for mask_method in CloudMaskMethod:
        masked_image = MaskedImage.from_id(image_id, mask_method=mask_method)
        masked_image._set_region_stats(region_10000ha)
        cl_portions.append(100 * masked_image.properties['CLOUDLESS_PORTION'])

    # test `mask_method` changes CLOUDLESS_PORTION but not by much
    assert len(set(cl_portions)) == len(cl_portions)
    assert all([cl_portions[0] != pytest.approx(cp, abs=10) for cp in cl_portions[1:]])
    assert all([cp != pytest.approx(0, abs=1) for cp in cl_portions])


@pytest.mark.parametrize('image_id', ['s2_sr_hm_image_id', 's2_toa_hm_image_id'])
def test_s2_cloudmask_mask_cirrus(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """Test S2 cloud/shadow masking `mask_cirrus` parameter with the `qa` method."""
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, mask_method='qa', mask_cirrus=False)
    # cloud and shadow free portion
    masked_image._set_region_stats(region_10000ha)
    non_cirrus_portion = 100 * masked_image.properties['CLOUDLESS_PORTION']
    masked_image = MaskedImage.from_id(image_id, mask_method='qa', mask_cirrus=True)
    masked_image._set_region_stats(region_10000ha)
    # cloud, cirrus and shadow free portion
    cirrus_portion = 100 * masked_image.properties['CLOUDLESS_PORTION']
    assert non_cirrus_portion >= cirrus_portion


@pytest.mark.parametrize('image_id', ['s2_sr_hm_image_id', 's2_toa_hm_image_id'])
def test_s2_cloudmask_dark(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """Test S2 cloud/shadow masking `dark` parameter."""
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, mask_method='cloud-prob', dark=0.5)
    masked_image._set_region_stats(region_10000ha)
    dark_pt5_portion = 100 * masked_image.properties['CLOUDLESS_PORTION']
    masked_image = MaskedImage.from_id(image_id, mask_method='cloud-prob', dark=0.1)
    masked_image._set_region_stats(region_10000ha)
    dark_pt1_portion = 100 * masked_image.properties['CLOUDLESS_PORTION']
    # test that increasing `dark` results in an increase in detected shadow and corresponding decrease in
    # CLOUDLESS_PORTION
    assert dark_pt1_portion > dark_pt5_portion


@pytest.mark.parametrize('image_id', ['s2_sr_hm_image_id', 's2_toa_hm_image_id'])
def test_s2_cloudmask_shadow_dist(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """Test S2 cloud/shadow masking `shadow_dist` parameter."""
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, mask_method='cloud-prob', shadow_dist=200)
    masked_image._set_region_stats(region_10000ha)
    sd200_portion = 100 * masked_image.properties['CLOUDLESS_PORTION']
    masked_image = MaskedImage.from_id(image_id, mask_method='cloud-prob', shadow_dist=1000)
    masked_image._set_region_stats(region_10000ha)
    sd1000_portion = 100 * masked_image.properties['CLOUDLESS_PORTION']
    # test that increasing `shadow_dist` results in an increase in detected shadow and corresponding decrease in
    # CLOUDLESS_PORTION
    assert sd200_portion > sd1000_portion


@pytest.mark.parametrize('image_id', ['s2_sr_hm_image_id', 's2_toa_hm_image_id'])
def test_s2_cloudmask_cdi_thresh(image_id: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """Test S2 cloud/shadow masking `cdi_thresh` parameter."""
    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, mask_method='cloud-prob', cdi_thresh=0.5)
    masked_image._set_region_stats(region_10000ha)
    cdi_pt5_portion = 100 * masked_image.properties['CLOUDLESS_PORTION']
    masked_image = MaskedImage.from_id(image_id, mask_method='cloud-prob', cdi_thresh=-0.5)
    masked_image._set_region_stats(region_10000ha)
    cdi_negpt5_portion = 100 * masked_image.properties['CLOUDLESS_PORTION']
    # test that increasing `cdi_thresh` results in an increase in detected cloud and corresponding decrease in
    # CLOUDLESS_PORTION
    assert cdi_negpt5_portion > cdi_pt5_portion


@pytest.mark.parametrize('image_id, max_cloud_dist', [('s2_sr_hm_image_id', 100), ('s2_sr_hm_image_id', 400)])
def test_s2_cloud_dist_max(image_id: str, max_cloud_dist: int, region_10000ha: Dict, request: pytest.FixtureRequest):
    """Test S2 cloud distance `max_cloud_dist` parameter."""

    def get_max_cloud_dist(cloud_dist: ee.Image):
        """Get the maximum of `cloud_dist` over region_10000ha."""
        mcd = cloud_dist.reduceRegion(reducer='max', geometry=region_10000ha, bestEffort=True, maxPixels=1e4)
        return mcd.get('CLOUD_DIST').getInfo() * 10

    image_id: str = request.getfixturevalue(image_id)
    masked_image = MaskedImage.from_id(image_id, max_cloud_dist=max_cloud_dist, mask_method='cloud-score')
    cloud_dist = masked_image.ee_image.select('CLOUD_DIST')
    meas_max_cloud_dist = get_max_cloud_dist(cloud_dist)
    assert meas_max_cloud_dist == pytest.approx(max_cloud_dist, rel=0.1)


@pytest.mark.parametrize(
    'masked_image',
    ['s2_sr_hm_qa_zero_masked_image', 's2_sr_hm_nocp_masked_image', 's2_sr_hm_nocs_masked_image'],
)
def test_s2_region_stats_missing_data(masked_image: str, region_10000ha: dict, request: pytest.FixtureRequest):
    """Test S2 region stats for unmasked images missing required cloud data."""
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    masked_image._set_region_stats(region_10000ha, scale=60)

    assert masked_image.properties is not None
    assert masked_image.properties['CLOUDLESS_PORTION'] == pytest.approx(0, abs=1)
    assert masked_image.properties['FILL_PORTION'] == pytest.approx(100, abs=1)


def _test_aux_stats(masked_image: MaskedImage, region: Dict):
    """Sanity tests on cloud/shadow etc. aux bands for a given MaskedImage and region."""
    # create a pan image for testing brightnesses
    ee_image = masked_image.ee_image
    pan = ee_image.select([0, 1, 2]).reduce(ee.Reducer.mean())

    # mask the pan image with cloud, shadow & cloudless masks
    cloud_mask = ee_image.select('CLOUD_MASK')
    shadow_mask = ee_image.select('SHADOW_MASK')
    cloudless_mask = ee_image.select('CLOUDLESS_MASK')
    pan_cloud = pan.updateMask(cloud_mask).rename('PAN_CLOUD')
    pan_shadow = pan.updateMask(shadow_mask).rename('PAN_SHADOW')
    pan_cloudless = pan.updateMask(cloudless_mask).rename('PAN_CLOUDLESS')

    # mask the cloud distance image with cloud & cloudless masks
    cdist = ee_image.select('CLOUD_DIST')
    cdist_cloud = cdist.updateMask(cloud_mask).rename('CDIST_CLOUD')
    cdist_cloudless = cdist.updateMask(cloudless_mask).rename('CDIST_CLOUDLESS')

    # find mean stats of all masked images, and min of cloud distance where it is >0
    stats_image = ee.Image([pan_cloud, pan_shadow, pan_cloudless, cdist_cloud, cdist_cloudless])
    proj = masked_image._ee_proj
    stats = stats_image.reduceRegion(
        reducer='mean', geometry=region, crs=proj, scale=proj.nominalScale(), bestEffort=True, maxPixels=1e8
    )
    cdist_min = cdist.updateMask(cdist).reduceRegion(
        reducer='min', geometry=region, crs=proj, scale=proj.nominalScale(), bestEffort=True, maxPixels=1e8
    )
    stats = stats.set('CDIST_MIN', cdist_min.get('CLOUD_DIST'))
    stats = stats.getInfo()

    # test cloud is brighter than cloudless
    assert stats['PAN_CLOUD'] > stats['PAN_CLOUDLESS']
    # test cloudless is brighter than shadow
    assert stats['PAN_CLOUDLESS'] > stats['PAN_SHADOW']
    # test cloudless areas have greater distance to cloud than cloudy areas
    assert stats['CDIST_CLOUDLESS'] > stats['CDIST_CLOUD']
    # test min distance to cloud is pixel size
    assert stats['CDIST_MIN'] * 10 == masked_image._ee_proj.nominalScale().getInfo()


@pytest.mark.parametrize(
    'masked_image', ['l9_masked_image', 'l8_masked_image', 'l7_masked_image', 'l5_masked_image', 'l4_masked_image']
)
def test_landsat_aux_bands(masked_image: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """Test Landsat auxiliary band values for sanity."""
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    _test_aux_stats(masked_image, region_10000ha)


@pytest.mark.parametrize(
    'image_id, mask_methods',
    [
        ('s2_sr_image_id', ['cloud-prob', 'qa']),
        ('s2_toa_image_id', ['cloud-prob', 'qa']),
        ('s2_sr_hm_image_id', ['cloud-prob', 'qa']),
        ('s2_toa_hm_image_id', ['cloud-prob', 'qa']),
        # missing QA60 so do cloud-prob method only
        ('s2_sr_hm_qa_zero_image_id', ['cloud-prob']),
    ],
)
def test_s2_aux_bands(image_id: str, mask_methods: Iterable, region_10000ha: Dict, request: pytest.FixtureRequest):
    """Test Sentinel-2 auxiliary band values for sanity with all masking methods."""
    image_id: str = request.getfixturevalue(image_id)
    for mask_method in mask_methods:
        masked_image = MaskedImage.from_id(image_id, mask_method=mask_method)
        _test_aux_stats(masked_image, region_10000ha)


@pytest.mark.parametrize(
    'masked_image',
    ['s2_sr_hm_nocp_masked_image', 's2_sr_hm_qa_zero_masked_image', 's2_sr_hm_nocs_masked_image'],
)
def test_s2_aux_bands_missing_data(masked_image: str, region_10000ha: Dict, request: pytest.FixtureRequest):
    """Test Sentinel-2 auxiliary band masking / transparency for unmasked images missing required cloud data."""
    masked_image: MaskedImage = request.getfixturevalue(masked_image)

    # get region sums of the auxiliary masks
    proj = masked_image._ee_proj
    aux_bands = masked_image.ee_image.select('.*MASK|CLOUD_DIST')
    aux_mask = aux_bands.mask()
    stats = aux_mask.reduceRegion(
        reducer='sum', geometry=region_10000ha, crs=proj, scale=proj.nominalScale(), bestEffort=True, maxPixels=1e8
    )
    stats = stats.getInfo()

    # test auxiliary masks are transparent
    assert stats['FILL_MASK'] > 0
    # s2_sr_hm_nocs_masked_image is missing CLOUD_MASK and SHADOW_MASK bands, so only include these when they exist
    band_names = ['CLOUDLESS_MASK', 'CLOUD_DIST'] + list({'CLOUD_MASK', 'SHADOW_MASK'}.intersection(stats.keys()))
    for band_name in band_names:
        assert stats[band_name] == 0, band_name


@pytest.mark.parametrize(
    'masked_image',
    [
        'gedi_cth_masked_image',
        # use s2_sr_masked_image rather than s2_sr_hm_masked_image which complicates testing due to fully masked
        # MSK_CLASSI* bands
        's2_sr_masked_image',
        'l9_masked_image',
    ],
)
def test_mask_clouds(masked_image: str, region_100ha: Dict, tmp_path, request: pytest.FixtureRequest):
    """Test MaskedImage.mask_clouds() masks the fill or cloudless portion by downloading and examining dataset masks."""
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    filename = tmp_path.joinpath(f'test_image.tif')
    masked_image.mask_clouds()
    proj_scale = masked_image._ee_proj.nominalScale().getInfo()
    masked_image.download(filename, region=region_100ha, dtype='float32', scale=proj_scale)
    assert filename.exists()

    with rio.open(filename, 'r') as ds:
        mask_name = 'CLOUDLESS_MASK' if isinstance(masked_image, CloudMaskedImage) else 'FILL_MASK'
        mask = ds.read(ds.descriptions.index(mask_name) + 1, masked=True)

        # test areas outside CLOUDLESS_MASK / FILL_MASK are masked
        assert np.all(mask == 1)

        # test CLOUDLESS_MASK / FILL_MASK matches the nodata mask for each band
        mask = mask.filled(0).astype('bool')
        ds_masks = ds.read_masks().astype('bool')
        assert np.all(mask == ds_masks)


@pytest.mark.parametrize(
    'masked_image',
    ['s2_sr_hm_nocp_masked_image', 's2_sr_hm_qa_zero_masked_image', 's2_sr_hm_nocs_masked_image'],
)
def test_s2_mask_clouds_missing_data(masked_image: str, region_100ha: Dict, tmp_path, request: pytest.FixtureRequest):
    """Test Sentinel2SrClImage.mask_clouds() masks the entire image when it is missing required cloud data. Downloads
    and examines dataset masks.
    """
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    filename = tmp_path.joinpath(f'test_image.tif')
    masked_image.mask_clouds()
    proj_scale = masked_image._ee_proj.nominalScale().getInfo()
    masked_image.download(filename, region=region_100ha, dtype='float32', scale=proj_scale)
    assert filename.exists()

    # test all data is masked / nodata
    with rio.open(filename, 'r') as ds:
        ds_masks = ds.read_masks().astype('bool')
        assert not np.any(ds_masks)


def test_skysat_region_stats():
    """Test _set_region_stats() works on SKYSAT image with no region."""
    ee_image = ee.Image('SKYSAT/GEN-A/PUBLIC/ORTHO/RGB/s02_20141004T074858Z')
    masked_image = MaskedImage(ee_image)
    masked_image._set_region_stats()
    assert 'FILL_PORTION' in masked_image.properties
    assert masked_image.properties['FILL_PORTION'] > 0.8
